# Output backends for sigmac
# Copyright 2016-2018 Thomas Patzke, Florian Roth

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import sys

import sigma
import yaml
import re

from sigma.backends.exceptions import NotSupportedError
from sigma.parser.condition import ConditionOR, NodeSubexpression
from .mixins import RulenameCommentMixin, QuoteCharMixin
from sigma.parser.modifiers.base import SigmaTypeModifier

class BackendOptions(dict):
    """
    Object containing all the options that should be passed to the backend.

    The options can come from command line and a YAML configuration file, and will be merged together.
    Options from the command line take precedence.
    """

    def __init__(self, options, config_file):
        """
        :param options: unparsed options coming from the CLI
        :param config_file: path to a YAML configuration file
        """

        self._load_config_file(config_file)
        self._parse_options(options)

    def _parse_options(self, options):
        """
        Populates options from the unparsed options of the CLI

        :param options: list unparsed options from the CLI.
            Each option can have one of the following formats:
            - "key=value": the option key:value will be passed to the backend
            - "key": the option key:True will be passed to the backend
        """

        if options is None:
            return

        for option in options:
            parsed = option.split("=", 1)
            try:
                self[parsed[0]] = parsed[1]
            except IndexError:
                # If the option is present but doesn't map to a value, treat it as a boolean flag
                self[parsed[0]] = True

    def _load_config_file(self, path):
        """
        Populates options from a configuration file

        :param path: Path to the configuration file
        """
        if path is None:
            return

        try:
            with open(path, 'r') as config_file:
                backend_config = yaml.safe_load(config_file.read())
                self.update(backend_config)
        except (IOError, OSError) as e:
            print("Failed to open backend configuration file '%s': %s" % (path, str(e)), file=sys.stderr)
            exit(1)
        except yaml.YAMLError as e:
            print("Failed to parse backend configuration file '%s' as valid YAML: %s" % (path, str(e)), file=sys.stderr)
            exit(1)

### Generic backend base classes
class BaseBackend:
    """Base class for all backends"""
    identifier = "base"
    active = False
    index_field = None    # field name that is used to address indices
    file_list = None
    options = tuple()     # a list of tuples with following elements: option name, default value, help text, target attribute name (option name if None)
    config_required = True
    default_config = None
    mapExpression = ""
    ymlFileName = None

    def __init__(self, sigmaconfig, backend_options=dict()):
        """
        Initialize backend. This gets a sigmaconfig object, which is notified about the used backend class by
        passing the object instance to it.
        """
        super().__init__()
        if not isinstance(sigmaconfig, (sigma.configuration.SigmaConfiguration, sigma.configuration.SigmaConfigurationChain, None)):
            raise TypeError("SigmaConfiguration object expected")
        self.backend_options = backend_options
        self.sigmaconfig = sigmaconfig
        self.sigmaconfig.set_backend(self)

        # Parse options
        for option, default_value, _, target in self.options:
            if target is None:
                target = option
            setattr(self, target, self.backend_options.setdefault(option, default_value))

    def setYmlFileName(self, filename):
        self.ymlFileName = filename

    def generate(self, sigmaparser):
        """Method is called for each sigma rule and receives the parsed rule (SigmaParser)"""
        if len(sigmaparser.condparsed) > 1:
            raise NotImplementedError("Base backend doesn't support multiple conditions")
        for parsed in sigmaparser.condparsed:
            query = self.generateQuery(parsed)
            before = self.generateBefore(parsed)
            after = self.generateAfter(parsed)

            result = ""
            if before is not None:
                result = before
            if query is not None:
                result += query
            if after is not None:
                result += after

            return result

    def generateQuery(self, parsed):
        result = self.generateNode(parsed.parsedSearch)
        if parsed.parsedAgg:
            result += self.generateAggregation(parsed.parsedAgg)
        #result = self.applyOverrides(result)
        return result

    def applyOverrides(self, query):
        try:
            if 'overrides' in self.sigmaconfig.config and isinstance(query, str):
                for expression in self.sigmaconfig.config['overrides']:
                    if 'regexes' in expression:
                        for x in expression['regexes']:
                            sub = expression['field']
                            value = expression['value']
                            query = re.sub(x, self.mapExpression % (sub, value), query)
                    if 'literals' in expression:
                        for x in expression['literals']:
                            sub = expression['field']
                            value = expression['value']
                            query = query.replace(x, self.mapExpression % (sub, value))
        except Exception:
            pass
        return query

    def generateNode(self, node):
        if type(node) == sigma.parser.condition.ConditionAND:
            return self.applyOverrides(self.generateANDNode(node))
        elif type(node) == sigma.parser.condition.ConditionOR:
            return self.applyOverrides(self.generateORNode(node))
        elif type(node) == sigma.parser.condition.ConditionNOT:
            return self.applyOverrides(self.generateNOTNode(node))
        elif type(node) == sigma.parser.condition.ConditionNULLValue:
            return self.applyOverrides(self.generateNULLValueNode(node))
        elif type(node) == sigma.parser.condition.ConditionNotNULLValue:
            return self.applyOverrides(self.generateNotNULLValueNode(node))
        elif type(node) == sigma.parser.condition.NodeSubexpression:
            return self.applyOverrides(self.generateSubexpressionNode(node))
        elif type(node) == sigma.parser.condition.SigmaSearchValueAsIs:
            return self.generateValueAsIsNode(node)
        elif type(node) == tuple:
            return self.applyOverrides(self.generateMapItemNode(node))
        elif type(node) in (str, int):
            return self.applyOverrides(self.generateValueNode(node))
        elif type(node) == list:
            return self.applyOverrides(self.generateListNode(node))
        elif isinstance(node, SigmaTypeModifier):
            return self.applyOverrides(self.generateTypedValueNode(node))
        else:
            raise TypeError("Node type %s was not expected in Sigma parse tree" % (str(type(node))))

    def generateValueAsIsNode(self, node):
        raise NotImplementedError("Node type not implemented for this backend")

    def generateANDNode(self, node):
        raise NotImplementedError("Node type not implemented for this backend")

    def generateORNode(self, node):
        raise NotImplementedError("Node type not implemented for this backend")

    def generateNOTNode(self, node):
        raise NotImplementedError("Node type not implemented for this backend")

    def generateSubexpressionNode(self, node):
        raise NotImplementedError("Node type not implemented for this backend")

    def generateListNode(self, node):
        raise NotImplementedError("Node type not implemented for this backend")

    def generateMapItemNode(self, node):
        raise NotImplementedError("Node type not implemented for this backend")

    def generateValueNode(self, node):
        raise NotImplementedError("Node type not implemented for this backend")

    def generateTypedValueNode(self, node):
        raise NotImplementedError("Node type not implemented for this backend")

    def generateNULLValueNode(self, node):
        raise NotImplementedError("Node type not implemented for this backend")

    def generateNotNULLValueNode(self, node):
        raise NotImplementedError("Node type not implemented for this backend")

    def generateAggregation(self, agg):
        raise NotImplementedError("Aggregations not implemented for this backend")

    def generateBefore(self, parsed):
        return ""

    def generateAfter(self, parsed):
        return ""

    def initialize(self):
        """
        Is called before the first file was processed with generate(). The right place if this backend is not intended to
        look isolated at each rule, but generates an output which incorporates multiple rules, e.g. dashboards.
        """
        pass

    def finalize(self):
        """
        Is called after the last file was processed with generate(). The right place if this backend is not intended to
        look isolated at each rule, but generates an output which incorporates multiple rules, e.g. dashboards.
        """
        pass

class SingleTextQueryBackend(RulenameCommentMixin, BaseBackend, QuoteCharMixin):
    """Base class for backends that generate one text-based expression from a Sigma rule"""
    identifier = "base-textquery"
    active = False

    # the following class variables define the generation and behavior of queries from a parse tree some are prefilled with default values that are quite usual
    andToken = None                     # Token used for linking expressions with logical AND
    orToken = None                      # Same for OR
    notToken = None                     # Same for NOT
    subExpression = None                # Syntax for subexpressions, usually parenthesis around it. %s is inner expression
    listExpression = None               # Syntax for lists, %s are list items separated with listSeparator
    listSeparator = None                # Character for separation of list items
    valueExpression = None              # Expression of values, %s represents value
    typedValueExpression = dict()       # Expression of typed values generated by type modifiers. modifier identifier -> expression dict, %s represents value
    nullExpression = None               # Expression of queries for null values or non-existing fields. %s is field name
    notNullExpression = None            # Expression of queries for not null values. %s is field name
    mapExpression = None                # Syntax for field/value conditions. First %s is fieldname, second is value
    mapListsSpecialHandling = False     # Same handling for map items with list values as for normal values (strings, integers) if True, generateMapItemListNode method is called with node
    mapListValueExpression = None       # Syntax for field/value condititons where map value is a list

    sort_condition_lists = False        # Sort condition items for AND and OR conditions

    def generateValueAsIsNode(self, node):
        if type(node.value) is list:
            return self.listExpression % (self.listSeparator.join(node.value))
        return self.listExpression % node.value

    def generateANDNode(self, node):
        generated = [ self.generateNode(val) for val in node ]
        filtered = [ g for g in generated if g is not None ]
        if filtered:
            if self.sort_condition_lists:
                filtered = sorted(filtered)
            return self.andToken.join(filtered)
        else:
            return None

    def generateORNode(self, node):
        generated = [ self.generateNode(val) for val in node ]
        filtered = [ g for g in generated if g is not None ]
        if filtered:
            if self.sort_condition_lists:
                filtered = sorted(filtered)
            return self.orToken.join(filtered)
        else:
            return None

    def generateNOTNode(self, node):
        generated = self.generateNode(node.item)
        if generated is not None:
            return self.notToken + generated
        else:
            return None

    def generateSubexpressionNode(self, node):
        generated = self.generateNode(node.items)
        if 'len'in dir(node.items): # fix the "TypeError: object of type 'NodeSubexpression' has no len()"
            if len(node.items) == 1:
                # A sub expression with length 1 is not a proper sub expression, no self.subExpression required
                return generated
        if generated:
            return self.subExpression % generated
        else:
            return None

    def generateListNode(self, node):
        if not set([type(value) for value in node]).issubset({str, int}):
            raise TypeError("List values must be strings or numbers")
        result = [self.generateNode(value) for value in node]
        if len(result) == 1:
            # A list with length 1 is not a proper list, no self.listExpression required
            return result[0]
        return self.listExpression % (self.listSeparator.join(result))

    def generateMapItemNode(self, node):
        fieldname, value = node

        transformed_fieldname = self.fieldNameMapping(fieldname, value)
        if self.mapListsSpecialHandling == False and type(value) in (str, int, list) or self.mapListsSpecialHandling == True and type(value) in (str, int):
            return self.mapExpression % (transformed_fieldname, self.generateNode(value))
        elif type(value) == list:
            return self.generateMapItemListNode(transformed_fieldname, value)
        elif isinstance(value, SigmaTypeModifier):
            return self.generateMapItemTypedNode(transformed_fieldname, value)
        elif value is None:
            return self.nullExpression % (transformed_fieldname, )
        elif isinstance(value, NodeSubexpression):
            return self.generateSubexpressionNode(value)
        else:
            raise TypeError("Backend does not support map values of type " + str(type(value)))

    def generateMapItemListNode(self, fieldname, value):
        return self.mapListValueExpression % (fieldname, self.generateNode(value))

    def generateMapItemTypedNode(self, fieldname, value):
        return self.mapExpression % (fieldname, self.generateTypedValueNode(value))

    def generateValueNode(self, node):
        return self.valueExpression % (self.cleanValue(str(node)))

    def generateTypedValueNode(self, node):
        try:
            return self.typedValueExpression[type(node)] % (str(node))
        except KeyError:
            raise NotImplementedError("Type modifier '{}' is not supported by backend".format(node.identifier))

    def generateNULLValueNode(self, node):
        return self.nullExpression % (node.item)

    def generateNotNULLValueNode(self, node):
        return self.notNullExpression % (node.item)

    def fieldNameMapping(self, fieldname, value):
        """
        Alter field names depending on the value(s). Backends may use this method to perform a final transformation of the field name
        in addition to the field mapping defined in the conversion configuration. The field name passed to this method was already
        transformed from the original name given in the Sigma rule.
        """
        return fieldname



# Static data required by backends

sysmon_schema = """
<manifest schemaversion="4.2" binaryversion="8.00">
  <configuration>
    <options>
      <!-- Command-line only options -->
      <option switch="i" name="Install" argument="optional" noconfig="true" exclusive="true" />
      <option switch="c" name="Configuration" argument="optional" noconfig="true" exclusive="true" />
      <option switch="u" name="UnInstall" argument="none" noconfig="true" exclusive="true" />
      <option switch="m" name="Manifest" argument="none" noconfig="true" exclusive="true" />
      <option switch="t" name="DebugMode" argument="none" noconfig="true" />
      <option switch="s" name="PrintSchema" argument="optional" noconfig="true" exclusive="true" />
      <option switch="nologo" name="NoLogo" argument="none" noconfig="true" />
      <option switch="accepteula" name="AcceptEula" argument="none" noconfig="true" />
      <option switch="-" name="ConfigDefault" argument="none" noconfig="true" />
      <!-- Configuration file -->
      <option switch="h" name="HashAlgorithms" argument="required" />
      <option switch="n" name="NetworkConnect" argument="optional" rule="true" />
      <option switch="l" name="ImageLoad" argument="optional" rule="true" />
      <option switch="d" name="DriverName" argument="required" />
      <option switch="k" name="ProcessAccess" argument="required" rule="true" forceconfig="true" />
      <option switch="r" name="CheckRevocation" argument="none" />
      <option switch="g" name="PipeMonitoring" argument="required" rule="true" forceconfig="true" />
    </options>
    <filters default="is">is,is not,contains,excludes,begin with,end with,less than,more than,image</filters>
  </configuration>
  <events>
    <event name="SYSMON_ERROR" value="255" level="Error" template="Error report" version="3">
      <data name="UtcTime" inType="win:UnicodeString" outType="xs:string" />
      <data name="ID" inType="win:UnicodeString" outType="xs:string" />
      <data name="Description" inType="win:UnicodeString" outType="xs:string" />
    </event>
    <event name="SYSMON_CREATE_PROCESS" value="1" level="Informational" template="Process Create" rulename="ProcessCreate" ruledefault="include" version="5">
      <data name="RuleName" inType="win:UnicodeString" outType="xs:string" />
      <data name="UtcTime" inType="win:UnicodeString" outType="xs:string" />
      <data name="ProcessGuid" inType="win:GUID" />
      <data name="ProcessId" inType="win:UInt32" outType="win:PID" />
      <data name="Image" inType="win:UnicodeString" outType="xs:string" />
      <data name="FileVersion" inType="win:UnicodeString" outType="xs:string" />
      <data name="Description" inType="win:UnicodeString" outType="xs:string" />
      <data name="Product" inType="win:UnicodeString" outType="xs:string" />
      <data name="Company" inType="win:UnicodeString" outType="xs:string" />
      <data name="CommandLine" inType="win:UnicodeString" outType="xs:string" />
      <data name="CurrentDirectory" inType="win:UnicodeString" outType="xs:string" />
      <data name="User" inType="win:UnicodeString" outType="xs:string" />
      <data name="LogonGuid" inType="win:GUID" />
      <data name="LogonId" inType="win:HexInt64" />
      <data name="TerminalSessionId" inType="win:UInt32" />
      <data name="IntegrityLevel" inType="win:UnicodeString" outType="xs:string" />
      <data name="Hashes" inType="win:UnicodeString" outType="xs:string" />
      <data name="ParentProcessGuid" inType="win:GUID" />
      <data name="ParentProcessId" inType="win:UInt32" outType="win:PID" />
      <data name="ParentImage" inType="win:UnicodeString" outType="xs:string" />
      <data name="ParentCommandLine" inType="win:UnicodeString" outType="xs:string" />
    </event>
    <event name="SYSMON_FILE_TIME" value="2" level="Informational" template="File creation time changed" rulename="FileCreateTime" ruledefault="include" version="4">
      <data name="RuleName" inType="win:UnicodeString" outType="xs:string" />
      <data name="UtcTime" inType="win:UnicodeString" outType="xs:string" />
      <data name="ProcessGuid" inType="win:GUID" />
      <data name="ProcessId" inType="win:UInt32" outType="win:PID" />
      <data name="Image" inType="win:UnicodeString" outType="xs:string" />
      <data name="TargetFilename" inType="win:UnicodeString" outType="xs:string" />
      <data name="CreationUtcTime" inType="win:UnicodeString" outType="xs:string" />
      <data name="PreviousCreationUtcTime" inType="win:UnicodeString" outType="xs:string" />
    </event>
    <event name="SYSMON_NETWORK_CONNECT" value="3" level="Informational" template="Network connection detected" rulename="NetworkConnect" version="5">
      <data name="RuleName" inType="win:UnicodeString" outType="xs:string" />
      <data name="UtcTime" inType="win:UnicodeString" outType="xs:string" />
      <data name="ProcessGuid" inType="win:GUID" />
      <data name="ProcessId" inType="win:UInt32" outType="win:PID" />
      <data name="Image" inType="win:UnicodeString" outType="xs:string" />
      <data name="User" inType="win:UnicodeString" outType="xs:string" />
      <data name="Protocol" inType="win:UnicodeString" outType="xs:string" />
      <data name="Initiated" inType="win:Boolean" />
      <data name="SourceIsIpv6" inType="win:Boolean" />
      <data name="SourceIp" inType="win:UnicodeString" outType="xs:string" />
      <data name="SourceHostname" inType="win:UnicodeString" outType="xs:string" />
      <data name="SourcePort" inType="win:UInt16" />
      <data name="SourcePortName" inType="win:UnicodeString" outType="xs:string" />
      <data name="DestinationIsIpv6" inType="win:Boolean" />
      <data name="DestinationIp" inType="win:UnicodeString" outType="xs:string" />
      <data name="DestinationHostname" inType="win:UnicodeString" outType="xs:string" />
      <data name="DestinationPort" inType="win:UInt16" />
      <data name="DestinationPortName" inType="win:UnicodeString" outType="xs:string" />
    </event>
    <event name="SYSMON_SERVICE_STATE_CHANGE" value="4" level="Informational" template="Sysmon service state changed" version="3">
      <data name="UtcTime" inType="win:UnicodeString" outType="xs:string" />
      <data name="State" inType="win:UnicodeString" outType="xs:string" />
      <data name="Version" inType="win:UnicodeString" outType="xs:string" />
      <data name="SchemaVersion" inType="win:UnicodeString" outType="xs:string" />
    </event>
    <event name="SYSMON_PROCESS_TERMINATE" value="5" level="Informational" template="Process terminated" rulename="ProcessTerminate" ruledefault="include" version="3">
      <data name="RuleName" inType="win:UnicodeString" outType="xs:string" />
      <data name="UtcTime" inType="win:UnicodeString" outType="xs:string" />
      <data name="ProcessGuid" inType="win:GUID" />
      <data name="ProcessId" inType="win:UInt32" outType="win:PID" />
      <data name="Image" inType="win:UnicodeString" outType="xs:string" />
    </event>
    <event name="SYSMON_DRIVER_LOAD" value="6" level="Informational" template="Driver loaded" rulename="DriverLoad" ruledefault="include" version="3">
      <data name="RuleName" inType="win:UnicodeString" outType="xs:string" />
      <data name="UtcTime" inType="win:UnicodeString" outType="xs:string" />
      <data name="ImageLoaded" inType="win:UnicodeString" outType="xs:string" />
      <data name="Hashes" inType="win:UnicodeString" outType="xs:string" />
      <data name="Signed" inType="win:UnicodeString" outType="xs:string" />
      <data name="Signature" inType="win:UnicodeString" outType="xs:string" />
      <data name="SignatureStatus" inType="win:UnicodeString" outType="xs:string" />
    </event>
    <event name="SYSMON_IMAGE_LOAD" value="7" level="Informational" template="Image loaded" rulename="ImageLoad" version="3">
      <data name="RuleName" inType="win:UnicodeString" outType="xs:string" />
      <data name="UtcTime" inType="win:UnicodeString" outType="xs:string" />
      <data name="ProcessGuid" inType="win:GUID" />
      <data name="ProcessId" inType="win:UInt32" outType="win:PID" />
      <data name="Image" inType="win:UnicodeString" outType="xs:string" />
      <data name="ImageLoaded" inType="win:UnicodeString" outType="xs:string" />
      <data name="FileVersion" inType="win:UnicodeString" outType="xs:string" />
      <data name="Description" inType="win:UnicodeString" outType="xs:string" />
      <data name="Product" inType="win:UnicodeString" outType="xs:string" />
      <data name="Company" inType="win:UnicodeString" outType="xs:string" />
      <data name="Hashes" inType="win:UnicodeString" outType="xs:string" />
      <data name="Signed" inType="win:UnicodeString" outType="xs:string" />
      <data name="Signature" inType="win:UnicodeString" outType="xs:string" />
      <data name="SignatureStatus" inType="win:UnicodeString" outType="xs:string" />
    </event>
    <event name="SYSMON_CREATE_REMOTE_THREAD" value="8" level="Informational" template="CreateRemoteThread detected" rulename="CreateRemoteThread" version="2">
      <data name="RuleName" inType="win:UnicodeString" outType="xs:string" />
      <data name="UtcTime" inType="win:UnicodeString" outType="xs:string" />
      <data name="SourceProcessGuid" inType="win:GUID" />
      <data name="SourceProcessId" inType="win:UInt32" outType="win:PID" />
      <data name="SourceImage" inType="win:UnicodeString" outType="xs:string" />
      <data name="TargetProcessGuid" inType="win:GUID" />
      <data name="TargetProcessId" inType="win:UInt32" outType="win:PID" />
      <data name="TargetImage" inType="win:UnicodeString" outType="xs:string" />
      <data name="NewThreadId" inType="win:UInt32" />
      <data name="StartAddress" inType="win:UnicodeString" outType="xs:string" />
      <data name="StartModule" inType="win:UnicodeString" outType="xs:string" />
      <data name="StartFunction" inType="win:UnicodeString" outType="xs:string" />
    </event>
    <event name="SYSMON_RAWACCESS_READ" value="9" level="Informational" template="RawAccessRead detected" rulename="RawAccessRead" version="2">
      <data name="RuleName" inType="win:UnicodeString" outType="xs:string" />
      <data name="UtcTime" inType="win:UnicodeString" outType="xs:string" />
      <data name="ProcessGuid" inType="win:GUID" />
      <data name="ProcessId" inType="win:UInt32" outType="win:PID" />
      <data name="Image" inType="win:UnicodeString" outType="xs:string" />
      <data name="Device" inType="win:UnicodeString" outType="xs:string" />
    </event>
    <event name="SYSMON_ACCESS_PROCESS" value="10" level="Informational" template="Process accessed" rulename="ProcessAccess" version="3">
      <data name="RuleName" inType="win:UnicodeString" outType="xs:string" />
      <data name="UtcTime" inType="win:UnicodeString" outType="xs:string" />
      <data name="SourceProcessGUID" inType="win:GUID" />
      <data name="SourceProcessId" inType="win:UInt32" outType="win:PID" />
      <data name="SourceThreadId" inType="win:UInt32" />
      <data name="SourceImage" inType="win:UnicodeString" outType="xs:string" />
      <data name="TargetProcessGUID" inType="win:GUID" />
      <data name="TargetProcessId" inType="win:UInt32" outType="win:PID" />
      <data name="TargetImage" inType="win:UnicodeString" outType="xs:string" />
      <data name="GrantedAccess" inType="win:HexInt32" />
      <data name="CallTrace" inType="win:UnicodeString" outType="xs:string" />
    </event>
    <event name="SYSMON_FILE_CREATE" value="11" level="Informational" template="File created" rulename="FileCreate" ruledefault="include" version="2">
      <data name="RuleName" inType="win:UnicodeString" outType="xs:string" />
      <data name="UtcTime" inType="win:UnicodeString" outType="xs:string" />
      <data name="ProcessGuid" inType="win:GUID" />
      <data name="ProcessId" inType="win:UInt32" outType="win:PID" />
      <data name="Image" inType="win:UnicodeString" outType="xs:string" />
      <data name="TargetFilename" inType="win:UnicodeString" outType="xs:string" />
      <data name="CreationUtcTime" inType="win:UnicodeString" outType="xs:string" />
    </event>
    <event name="SYSMON_REG_KEY" value="12" level="Informational" template="Registry object added or deleted" rulename="RegistryEvent" ruledefault="include" version="2">
      <data name="RuleName" inType="win:UnicodeString" outType="xs:string" />
      <data name="EventType" inType="win:UnicodeString" outType="xs:string" />
      <data name="UtcTime" inType="win:UnicodeString" outType="xs:string" />
      <data name="ProcessGuid" inType="win:GUID" />
      <data name="ProcessId" inType="win:UInt32" outType="win:PID" />
      <data name="Image" inType="win:UnicodeString" outType="xs:string" />
      <data name="TargetObject" inType="win:UnicodeString" outType="xs:string" />
    </event>
    <event name="SYSMON_REG_SETVALUE" value="13" level="Informational" template="Registry value set" rulename="RegistryEvent" ruledefault="include" version="2">
      <data name="RuleName" inType="win:UnicodeString" outType="xs:string" />
      <data name="EventType" inType="win:UnicodeString" outType="xs:string" />
      <data name="UtcTime" inType="win:UnicodeString" outType="xs:string" />
      <data name="ProcessGuid" inType="win:GUID" />
      <data name="ProcessId" inType="win:UInt32" outType="win:PID" />
      <data name="Image" inType="win:UnicodeString" outType="xs:string" />
      <data name="TargetObject" inType="win:UnicodeString" outType="xs:string" />
      <data name="Details" inType="win:UnicodeString" outType="xs:string" />
    </event>
    <event name="SYSMON_REG_NAME" value="14" level="Informational" template="Registry object renamed" rulename="RegistryEvent" ruledefault="include" version="2">
      <data name="RuleName" inType="win:UnicodeString" outType="xs:string" />
      <data name="EventType" inType="win:UnicodeString" outType="xs:string" />
      <data name="UtcTime" inType="win:UnicodeString" outType="xs:string" />
      <data name="ProcessGuid" inType="win:GUID" />
      <data name="ProcessId" inType="win:UInt32" outType="win:PID" />
      <data name="Image" inType="win:UnicodeString" outType="xs:string" />
      <data name="TargetObject" inType="win:UnicodeString" outType="xs:string" />
      <data name="NewName" inType="win:UnicodeString" outType="xs:string" />
    </event>
    <event name="SYSMON_FILE_CREATE_STREAM_HASH" value="15" level="Informational" template="File stream created" rulename="FileCreateStreamHash" ruledefault="include" version="2">
      <data name="RuleName" inType="win:UnicodeString" outType="xs:string" />
      <data name="UtcTime" inType="win:UnicodeString" outType="xs:string" />
      <data name="ProcessGuid" inType="win:GUID" />
      <data name="ProcessId" inType="win:UInt32" outType="win:PID" />
      <data name="Image" inType="win:UnicodeString" outType="xs:string" />
      <data name="TargetFilename" inType="win:UnicodeString" outType="xs:string" />
      <data name="CreationUtcTime" inType="win:UnicodeString" outType="xs:string" />
      <data name="Hash" inType="win:UnicodeString" outType="xs:string" />
    </event>
    <event name="SYSMON_SERVICE_CONFIGURATION_CHANGE" value="16" level="Informational" template="Sysmon config state changed" version="3">
      <data name="UtcTime" inType="win:UnicodeString" outType="xs:string" />
      <data name="Configuration" inType="win:UnicodeString" outType="xs:string" />
      <data name="ConfigurationFileHash" inType="win:UnicodeString" outType="xs:string" />
    </event>
    <event name="SYSMON_CREATE_NAMEDPIPE" value="17" level="Informational" template="Pipe Created" rulename="PipeEvent" ruledefault="exclude" version="1">
      <data name="RuleName" inType="win:UnicodeString" outType="xs:string" />
      <data name="UtcTime" inType="win:UnicodeString" outType="xs:string" />
      <data name="ProcessGuid" inType="win:GUID" />
      <data name="ProcessId" inType="win:UInt32" outType="win:PID" />
      <data name="PipeName" inType="win:UnicodeString" outType="xs:string" />
      <data name="Image" inType="win:UnicodeString" outType="xs:string" />
    </event>
    <event name="SYSMON_CONNECT_NAMEDPIPE" value="18" level="Informational" template="Pipe Connected" rulename="PipeEvent" ruledefault="exclude" version="1">
      <data name="RuleName" inType="win:UnicodeString" outType="xs:string" />
      <data name="UtcTime" inType="win:UnicodeString" outType="xs:string" />
      <data name="ProcessGuid" inType="win:GUID" />
      <data name="ProcessId" inType="win:UInt32" outType="win:PID" />
      <data name="PipeName" inType="win:UnicodeString" outType="xs:string" />
      <data name="Image" inType="win:UnicodeString" outType="xs:string" />
    </event>
    <event name="SYSMON_WMI_FILTER" value="19" level="Informational" template="WmiEventFilter activity detected" rulename="WmiEvent" ruledefault="exclude" version="3">
      <data name="RuleName" inType="win:UnicodeString" outType="xs:string" />
      <data name="EventType" inType="win:UnicodeString" outType="xs:string" />
      <data name="UtcTime" inType="win:UnicodeString" outType="xs:string" />
      <data name="Operation" inType="win:UnicodeString" outType="xs:string" />
      <data name="User" inType="win:UnicodeString" outType="xs:string" />
      <data name="EventNamespace" inType="win:UnicodeString" outType="xs:string" />
      <data name="Name" inType="win:UnicodeString" outType="xs:string" />
      <data name="Query" inType="win:UnicodeString" outType="xs:string" />
    </event>
    <event name="SYSMON_WMI_CONSUMER" value="20" level="Informational" template="WmiEventConsumer activity detected" rulename="WmiEvent" ruledefault="exclude" version="3">
      <data name="RuleName" inType="win:UnicodeString" outType="xs:string" />
      <data name="EventType" inType="win:UnicodeString" outType="xs:string" />
      <data name="UtcTime" inType="win:UnicodeString" outType="xs:string" />
      <data name="Operation" inType="win:UnicodeString" outType="xs:string" />
      <data name="User" inType="win:UnicodeString" outType="xs:string" />
      <data name="Name" inType="win:UnicodeString" outType="xs:string" />
      <data name="Type" inType="win:UnicodeString" outType="xs:string" />
      <data name="Destination" inType="win:UnicodeString" outType="xs:string" />
    </event>
    <event name="SYSMON_WMI_BINDING" value="21" level="Informational" template="WmiEventConsumerToFilter activity detected" rulename="WmiEvent" ruledefault="exclude" version="3">
      <data name="RuleName" inType="win:UnicodeString" outType="xs:string" />
      <data name="EventType" inType="win:UnicodeString" outType="xs:string" />
      <data name="UtcTime" inType="win:UnicodeString" outType="xs:string" />
      <data name="Operation" inType="win:UnicodeString" outType="xs:string" />
      <data name="User" inType="win:UnicodeString" outType="xs:string" />
      <data name="Consumer" inType="win:UnicodeString" outType="xs:string" />
      <data name="Filter" inType="win:UnicodeString" outType="xs:string" />
    </event>
  </events>
</manifest>
"""



# Output backends for sigmac
# Copyright 2021 Datadog, Inc.

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from re import compile
from sigma.backends.base import SingleTextQueryBackend
from sigma.parser.condition import NodeSubexpression


class DatadogLogsBackend(SingleTextQueryBackend):
    """Converts Sigma rule into Datadog log search query."""

    identifier = "datadog-logs"
    active = True
    config_required = False

    andToken = " AND "
    orToken = " OR "
    notToken = "-"
    subExpression = "(%s)"
    listExpression = "(%s)"
    # List selection items are linked with a logical 'OR' per the Sigma specification:
    # https://github.com/SigmaHQ/sigma/wiki/Specification#lists.
    listSeparator = " OR "
    valueExpression = "%s"
    mapExpression = "%s:%s"
    nullExpression = "-%s:*"
    notNullExpression = "%s:*"

    # The escaped characters list comes from https://docs.datadoghq.com/logs/explorer/search_syntax/#escaping-of-special-characters.
    specialCharactersRegexp = compile(r'([+\-=&|><!(){}\[\]^"~?:\\/]+)')
    whitespacesRegexp = compile(r"\s")

    # Default tags taken from https://docs.datadoghq.com/getting_started/tagging/#introduction.
    tags = {}

    def __init__(self, sigmaconfig, backend_options=None):

        if sigmaconfig.config:
            self.dd_index = sigmaconfig.config.get("index")
            self.dd_service = sigmaconfig.config.get("service")
            self.dd_source = sigmaconfig.config.get("source")
            self.dd_env = sigmaconfig.config.get("env")
            self.tags = sigmaconfig.config.get("tags")

        # if backend_options defined, overwrite default config file (for retro-compatibility)
        if backend_options is None:
            backend_options = {}

        if "index" in backend_options:
            self.dd_index = backend_options["index"]

        if "service" in backend_options:
            self.dd_service = backend_options["service"]

        if "source" in backend_options:
            self.dd_source = backend_options["source"]

        if "env" in backend_options:
            self.dd_env = backend_options["env"]

        super().__init__(sigmaconfig)

    def generateQuery(self, parsed):
        nodes = []

        if hasattr(self, "dd_index") and self.dd_index != None:
            nodes.append(("index", self.dd_index))

        if hasattr(self, "dd_service") and self.dd_service != None:
            nodes.append(("service", self.dd_service))

        if hasattr(self, "dd_source") and self.dd_source != None:
            nodes.append(("source", self.dd_source))

        if hasattr(self, "dd_env") and self.dd_env != None:
            nodes.append(("env", self.dd_env))

        if type(parsed.parsedSearch) == NodeSubexpression:
            nodes.append(parsed.parsedSearch.items)
        else:
            nodes.append(parsed.parsedSearch)

        return self.generateANDNode(nodes)

    def cleanValue(self, val):
        if type(val) == int:
            return val
        else:
            # Whitespaces characters are replaced with a `?`.
            # Datadog also supports escaping whitespaces by double quoting
            # expression, but at the cost of losing the `*` pattern matching
            # syntax that we wanted to preserve.
            # Note that technically, `?` matches  **any** single character.
            return self.whitespacesRegexp.sub(
                # Special characters are escaped with a `\` which requires to be escaped
                # in Python as well (see https://docs.python.org/3/library/re.html).
                # This explains the unusual number of `\` in the following regex definition.
                "?", self.specialCharactersRegexp.sub("\\\\\g<1>", val)
            )

    def generateMapItemNode(self, node):
        key, value = node
        return super().generateMapItemNode(((self.wrap_key(key)), value))

    def generateNULLValueNode(self, node):
        return super().generateNULLValueNode((self.wrap_key(node)))

    def generateNotNULLValueNode(self, node):
        return super().generateNotNULLValueNode(self.wrap_key(node))

    def wrap_key(self, key):
        if key not in self.tags and key not in ['index', 'service', 'source', 'env']:
            # print("WARNING tag %s not converted" % format(key))
            return "@%s" % key
        elif key not in ['index', 'service', 'source', 'env']:
            return self.tags[key]
        else:
            return key



import re

import sigma
from sigma.backends.base import SingleTextQueryBackend
from sigma.backends.mixins import MultiRuleOutputMixin

from .exceptions import NotSupportedError


class SysmonConfigBackend(SingleTextQueryBackend, MultiRuleOutputMixin):
    """Converts Sigma rule into sysmon XML configuration"""
    identifier = "sysmon"
    active = True
    andToken = " AND "
    orToken = " OR "
    notToken = "NOT "
    subExpression = "(%s)"
    config_required = True
    INCLUDE = "include"
    EXCLUDE = "exclude"
    conditionDict = {
        "startswith": "begin with",
        "endswith": "end with",
        "all": "contains all"
    }

    def __init__(self, *args, **kwargs):
        self.table = None
        self.logsource = None
        self.allowedSource = {
            "process_creation": "ProcessCreate"
        }
        self.eventidTagMapping = {
            1: "ProcessCreate",
            4799: "ProcessCreate",
            2: "FileCreateTime",
            3: "NetworkConnect",
            5: "ProcessTerminate",
            6: "DriverLoad",
            7: "ImageLoad",
            8: "CreateRemoteThread",
            9: "RawAccessRead",
            10: "ProcessAccess",
            11: "FileCreate",
            12: "RegistryEvent",
            13: "RegistryEvent",
            14: "RegistryEvent",
            15: "FileCreateStreamHash",
            17: "PipeEvent",
            18: "PipeEvent",
            19: "WmiEvent",
            20: "WmiEvent",
            21: "WmiEvent",
            22: "DNSQuery",
            257: "DNSQuery",
            23: "FileDelete"
        }
        self.allowedCondCombinations = {
            'single': [
                [4],
                [1, 4],
                [2, 4],
            ],
            'multi': [
                [1, 2, 4],
            ],
            "exclude": [
                [1, 3, 4],
                [2, 3, 4]
            ],
            # "multi-exclude": [
            #     [1, 2, 3, 4]
            # ]
        }
        return super().__init__(*args, **kwargs)

    def cleanValue(self, value):
        val = re.sub("[*]", "", value)
        return val

    def mapFiledValue(self, field, value):
        condition = None
        any_selector = "contains any"
        if "|" in field:
            field, *pipes = field.split("|")
            if len(pipes) == 1:
                modifier = pipes[0]
                if modifier in self.conditionDict:
                    condition = self.conditionDict[modifier]
                if modifier == "all":
                    any_selector = "contains all"
            else:
                raise NotImplementedError("not implemented condition")
        if isinstance(value, list) and len(value) > 1:
            condition = any_selector
            value = ";".join(value)
        elif "*" in value:
            if value.startswith("*") and value.endswith("*"):
                condition = "contains"
            elif value.startswith("*"):
                condition = "end with"
            elif value.endswith("*"):
                condition = "begin with"
            else:
                condition = "contains"

        if condition:
            field_str = '<{field} condition="{condition}">{value}</{field}>'.format(field=field,
                                                                                    condition=condition,
                                                                                    value=self.cleanValue(value))
        else:
            field_str = '<{field}>{value}</{field}>'.format(field=field, value=self.cleanValue(value))

        return field_str

    def createRule(self, selections):
        fields_list = []
        table = None
        for field, value in selections.items():
            if isinstance(value, list) and len(value) == 1:
                value = value[0]
            if field == "EventID":
                try:
                    table = self.eventidTagMapping[value]
                except KeyError:
                    table = self.eventidTagMapping[1]
            else:
                created_field_value = self.mapFiledValue(field, value)
                fields_list.append(created_field_value)
        fields_list_filtered = [item for item in fields_list if item]
        if any(fields_list_filtered):
            rule = '''\n\t\t<Rule name="{rule_name}" groupRelation="and">\n\t\t\t{fields}\n\t\t</Rule>'''.format(rule_name=self.rule_name, fields="\n\t\t\t".join(["{}".format(item) for item in fields_list_filtered]))
            t = table if table else self.table
            return rule, t
        else:
            return None, None

    def createRuleGroup(self, condition_objects, condition, match_type="include"):
        rules = None
        rules_selections = [item for item in condition_objects if item.type == 4]
        if len(rules_selections) == 1:
            rule, table = self.createRule(self.detection.get(rules_selections[0].matched))
            rules = {match_type: {table: rule}}
        else:
            if "or" in condition.lower():
                result = {}
                for selection_object in rules_selections:
                    rule, table = self.createRule(self.detection.get(selection_object.matched))
                    if result.get(table):
                        result[table].append(rule)
                    else:
                        result[table] = [rule]
                result = {table_name: "\n\t\t".join(rules_list) for table_name, rules_list in result.items()}
                rules = {match_type: result}
            elif "and" in condition.lower():
                rules_dict = {}
                for selection_object in rules_selections:
                    rules_dict.update(self.detection.get(selection_object.matched))
                rule, table = self.createRule(rules_dict)
                rules = {match_type: {table: rule}}
        if rules:
            rules_result = []
            for match, tables in rules.items():
                for table, rules in tables.items():
                    category_comment = '\n<!--Insert This Rule in <{} onmatch="{}"> section -->\n{}'.format(table, match,
                                                                                                            "".join(rules))
                    rules_result.append(category_comment)
            return "".join(rules_result)
        else:
            raise NotSupportedError("Couldn't create rule with current condition.")

    def createMultiRuleGroup(self, conditions):
        conditions_id = "".join([str(item.type) for item in conditions])
        or_index = conditions_id.index("2")
        sorted_conditions = [conditions[:or_index], conditions[or_index+1:]]
        if sorted_conditions:
            result = ""
            for rule_condition in sorted_conditions:
                rule = self.createRuleGroup(condition_objects=rule_condition, condition=" ".join([item.matched for item in rule_condition]))
                result += "{}\n".format(rule)
            return result
        else:
            raise NotSupportedError("Not implemented condition.")

    def createExcludeRuleGroup(self, conditions):
        conditions_id = "".join([str(item.type) for item in conditions])
        condition = self.detection.get("condition")
        sorted_conditions = None
        if "and not" in condition.lower():
            andnot_index = conditions_id.index("13")
            sorted_conditions = [(conditions[:andnot_index], self.INCLUDE), ([item for item in conditions if item.type != 3], self.EXCLUDE)]
        elif "or not" in condition.lower():
            ornot_index = conditions_id.index("23")
            sorted_conditions = [(conditions[:ornot_index], self.INCLUDE), (conditions[ornot_index + 2:], self.EXCLUDE)]
        if sorted_conditions:
            result = ""
            for rule_condition in sorted_conditions:
                rule = self.createRuleGroup(condition_objects=rule_condition[0], condition=" ".join([item.matched for item in rule_condition[0]]), match_type=rule_condition[1])
                result += "{}\n".format(rule)
            return result

    def checkRuleCondition(self, condtokens):
        if len(condtokens) == 1:
            conditions = [item for item in condtokens[0].tokens]
            conditions_combination = list(set([item.type for item in conditions]))
            for rule_type, combinations in self.allowedCondCombinations.items():
                for combination in combinations:
                    if sorted(conditions_combination) == sorted(combination):
                        return rule_type, conditions
            else:
                raise NotSupportedError("Not supported condition.")
        else:
            raise NotSupportedError("Not supported condition.")

    def createTableFromLogsource(self):
        if self.logsource.get("product", "") not in ("linux","windows"):
            raise NotSupportedError(
                "Not supported logsource. Should be product `linux` or `windows`.")
        for item in self.logsource.values():
            if str(item).lower() in self.allowedSource.keys():
                self.table = self.allowedSource.get(item.lower())
                break
        else:
            self.table = "ProcessCreate"

    def checkDetection(self):
        for selection_name, value in self.detection.items():
            if isinstance(value, list):
                raise NotSupportedError("Keywords are not supported in sysmon backend.")


    def generate(self, sigmaparser):
        sysmon_rule = None
        title = sigmaparser.parsedyaml.get("title", "")
        author = sigmaparser.parsedyaml.get("author", {})
        self.rule_name = "{} by {}".format(title, author)
        self.detection = sigmaparser.parsedyaml.get("detection", {})
        self.checkDetection()
        self.logsource = sigmaparser.parsedyaml["logsource"]
        self.createTableFromLogsource()
        rule_type, conditions = self.checkRuleCondition(sigmaparser.condtoken)
        if rule_type == "single":
            sysmon_rule = self.createRuleGroup(conditions, self.detection.get("condition"))
        elif rule_type == "multi":
            sysmon_rule = self.createMultiRuleGroup(conditions)
        elif rule_type == "exclude":
            sysmon_rule = self.createExcludeRuleGroup(conditions)

        if sysmon_rule:
            rulegroup_comment = '<!--RuleGroup groupRelation should be `or` <RuleGroup groupRelation="or"> -->'
            return "{}\n{}".format(rulegroup_comment, sysmon_rule)



#!/usr/bin/env python3
# A simple Sigma Configurations checker
# Copyright frack113

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import sigma.backends.discovery as backends
import ruamel.yaml
from pathlib import Path
from argparse import ArgumentParser
import sys
import csv
    
def main():
    argparser = ArgumentParser(description="A simple Sigma Configurations checker")
    argparser.add_argument("--verify", "-V", action="store_true", help="Verify if configuration file have valid backend name")
    argparser.add_argument("--sumary", "-s", action="store_true", help="Give some information.")
    argparser.add_argument("--error", "-e", action="store_true", help="Exit with error code 10 on verification failures.")
    argparser.add_argument("--output", "-o", default=None, help="Output csv file")
    args = argparser.parse_args()
    
    passed = True
    
    list_backend =[]
    for backend in sorted(backends.getBackendList(), key=lambda backend: backend.identifier):
        list_backend.append(backend.identifier)

    if args.sumary: 
        print(f"Backend found :\n{list_backend}\n")

    if args.verify:
        csv_lst = []
        valid = 0
        empty = 0
        faulty = 0
        yml_files =Path('config/').glob("*.yml")
        for yml in yml_files:
            print(f"Check configurations file : {yml.name}")
            with yml.open("r",encoding="UTF-8") as f:
                data = ruamel.yaml.load(f,Loader=ruamel.yaml.RoundTripLoader)
                if 'backends' in data:
                    for backend in data['backends']:
                        if backend in list_backend:
                            csv_lst.append([yml.name,backend,'OK'])
                            valid += 1
                        else:
                            csv_lst.append([yml.name,backend,'NOK'])
                            faulty += 1
                            passed = False
                else:
                    csv_lst.append([yml.name,"no backends section",'-'])
                    empty += 1
                    #passed = False 
                    #Should not be but not sure
            
        if args.sumary: 
            print('-------')
            print('Summary')
            print(f'Valid backend name: {valid}\nInvalid backend name: {faulty}\nFile with no Backend: {empty}')
            print('-------')

        if args.output:
            with open(args.output, 'w', newline='') as csvfile:
                spamwriter = csv.writer(csvfile, delimiter=';',quotechar='|', quoting=csv.QUOTE_MINIMAL)
                spamwriter.writerow(['Configurations Name','Backend Name','Result'])
                for row in csv_lst:
                    spamwriter.writerow(row)
                
        if not passed:
            print("**************************************")
            print("Some Configurations file are not valid")
            print("**************************************")
            if args.error:
                exit(10)
                

if __name__ == "__main__":
    main()



import sigma
from sigma.parser.modifiers.base import SigmaTypeModifier
from sigma.parser.modifiers.type import SigmaRegularExpressionModifier
from .base import SingleTextQueryBackend


class STIXBackend(SingleTextQueryBackend):
    """Converts Sigma rule into STIX pattern."""
    identifier = "stix"
    active = True
    andToken = " AND "
    orToken = " OR "
    notToken = "NOT "
    subExpression = "(%s)"
    valueExpression = "\'%s\'"
    mapExpression = "%s = %s"
    notMapExpression = "%s != %s"
    mapListsSpecialHandling = True
    sort_condition_lists = True

    def cleanKey(self, key):
        if key is None:
            raise TypeError("Backend does not support empty key " + str(key))
        else:
            return key

    def cleanValue(self, value):
        return value

    def generateANDNode(self, node, currently_within_NOT_node=False):
        generated = [self.generateNode(val, currently_within_NOT_node) for val in node]
        filtered = [g for g in generated if g is not None]
        if filtered:
            if self.sort_condition_lists:
                filtered = sorted(filtered)
            return self.andToken.join(filtered)
        else:
            return None

    def generateORNode(self, node, currently_within_NOT_node=False):
        generated = [self.generateNode(val, currently_within_NOT_node) for val in node]
        filtered = [g for g in generated if g is not None]
        if filtered:
            if self.sort_condition_lists:
                filtered = sorted(filtered)
            return self.orToken.join(filtered)
        else:
            return None

    def generateNOTNode(self, node, currently_within_NOT_node=False):
        currently_within_NOT_node = not(currently_within_NOT_node)
        generated = self.generateNode(node.item, currently_within_NOT_node)
        if generated is not None:
            return generated
        else:
            return None

    def generateSubexpressionNode(self, node, currently_within_NOT_node=False):
        generated = self.generateNode(node.items, currently_within_NOT_node)
        if generated:
            return self.subExpression % generated
        else:
            return None

    def generateMapItemNode(self, node, currently_within_NOT_node=False):
        fieldname, value = node

        transformed_fieldname = self.fieldNameMapping(fieldname, value)
        if self.mapListsSpecialHandling == False and type(value) in (str, int, list) or self.mapListsSpecialHandling == True and type(value) in (str, int):
            if currently_within_NOT_node:
                return self.notMapExpression % (transformed_fieldname, self.generateNode(value))
            return self.mapExpression % (transformed_fieldname, self.generateNode(value))
        elif type(value) == list:
            return self.generateMapItemListNode(transformed_fieldname, value, currently_within_NOT_node)
        elif isinstance(value, SigmaTypeModifier):
            return self.generateMapItemTypedNode(transformed_fieldname, value)
        elif value is None:
            return self.nullExpression % (transformed_fieldname, )
        else:
            raise TypeError("Backend does not support map values of type " + str(type(value)))

    def generateMapItemListNode(self, key, value, currently_within_NOT_node=False):
        items_list = list()
        for item in value:
            if type(item) == str and "*" in item:
                item = item.replace("*", "%")
                if currently_within_NOT_node:
                    items_list.append('%s NOT LIKE %s' % (self.cleanKey(key), self.generateValueNode(item)))
                else:
                    items_list.append('%s LIKE %s' % (self.cleanKey(key), self.generateValueNode(item)))
            else:
                if currently_within_NOT_node:
                    items_list.append('%s != %s' % (self.cleanKey(key), self.generateValueNode(item)))
                else:
                    items_list.append('%s = %s' % (self.cleanKey(key), self.generateValueNode(item)))
        if currently_within_NOT_node:
            return '(' + " AND ".join(items_list) + ')'
        else:
            return '('+" OR ".join(items_list)+')'

    def generateMapItemTypedNode(self, key, value):
        if type(value) == SigmaRegularExpressionModifier:
            regex = str(value)
            # Regular Expressions have to match the full value in QRadar
            if not (regex.startswith('^') or regex.startswith('.*')):
                regex = '.*' + regex
            if not (regex.endswith('$') or regex.endswith('.*')):
                regex = regex + '.*'
            return "%s MATCHES %s" % (self.cleanKey(key), self.generateValueNode(regex))
        else:
            raise NotImplementedError("Type modifier '{}' is not supported by backend".format(value.identifier))

    def generateMapItemNode(self, node, currently_within_NOT_node=False):
        key, value = node
        if ":" not in key:
            # key wasn't mapped
            return None
        if self.mapListsSpecialHandling == False and type(value) in (str, int, list) or self.mapListsSpecialHandling == True and type(value) in (str, int):
            if type(value) == str and "*" in value:
                value = value.replace("*", "%")
                if currently_within_NOT_node:
                    return "%s NOT LIKE %s" % (self.cleanKey(key), self.generateValueNode(value))
                return "%s LIKE %s" % (self.cleanKey(key), self.generateValueNode(value))
            elif type(value) in (str, int):
                if currently_within_NOT_node:
                    return self.notMapExpression % (self.cleanKey(key), self.generateValueNode(value))
                return self.mapExpression % (self.cleanKey(key), self.generateValueNode(value))
        elif type(value) == list:
            return self.generateMapItemListNode(key, value, currently_within_NOT_node)
        elif isinstance(value, SigmaTypeModifier):
            return self.generateMapItemTypedNode(key, value)
        else:
            raise TypeError("Backend does not support map values of type " + str(type(value)))

    def generateValueNode(self, node, keypresent=True):
        if keypresent == False:
            if type(node) == str and "*" in node:
                node = node.replace("*", "%")
            return "artifact:payload_bin LIKE \'{0}\'".format(self.cleanValue(str(node)))
        else:
            return self.valueExpression % (self.cleanValue(str(node)))

    def generateNode(self, node, currently_within_NOT_node=False):
        if type(node) == sigma.parser.condition.ConditionAND:
            if currently_within_NOT_node:
                return self.generateORNode(node, currently_within_NOT_node)
            return self.generateANDNode(node, currently_within_NOT_node)
        elif type(node) == sigma.parser.condition.ConditionOR:
            if currently_within_NOT_node:
                return self.generateANDNode(node, currently_within_NOT_node)
            return self.generateORNode(node, currently_within_NOT_node)
        elif type(node) == sigma.parser.condition.ConditionNOT:
            return self.generateNOTNode(node, currently_within_NOT_node)
        elif type(node) == sigma.parser.condition.NodeSubexpression:
            return self.generateSubexpressionNode(node, currently_within_NOT_node)
        elif type(node) == tuple:
            return self.generateMapItemNode(node, currently_within_NOT_node)
        elif type(node) in (str, int):
            return self.generateValueNode(node, keypresent=False)
        else:
            raise TypeError("Node type %s was not expected in Sigma parse tree" % (str(type(node))))

    def generate(self, sigmaparser):
        for parsed in sigmaparser.condparsed:
            query = self.generateQuery(parsed, sigmaparser)
            return "[" + query + "]"

    def generateQuery(self, parsed, sigmaparser):
        result = self.generateNode(parsed.parsedSearch)
        return result



# ArcSight backend for sigmac created by SOC Prime
# Copyright 2018 SOC Prime

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import re
from sigma.parser.condition import ConditionOR
from .base import SingleTextQueryBackend

class ArcSightBackend(SingleTextQueryBackend):
    """Converts Sigma rule into ArcSight saved search. Contributed by SOC Prime. https://socprime.com"""
    identifier = "arcsight"
    active = True
    andToken = " AND "
    orToken = " OR "
    notToken = " NOT "
    subExpression = "(%s)"
    listExpression = "(%s)"
    listSeparator = " OR "
    valueExpression = "\"%s\""
    containsExpression = "%s CONTAINS %s"
    nullExpression = "NOT _exists_:%s"
    notNullExpression = "_exists_:%s"
    mapExpression = "%s = %s"
    mapListsSpecialHandling = True
    mapListValueExpression = "%s = %s"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        aFL = ["deviceVendor", "categoryDeviceGroup", "deviceProduct"]
        for item in self.sigmaconfig.fieldmappings.values():
            if item.target_type is list:
                aFL.extend(item.target)
            else:
                aFL.append(item.target)
        self.allowedFieldsList = list(set(aFL))

    # Skip logsource value from sigma document for separate path.
    def generateCleanValueNodeLogsource(self, value):
        return self.valueExpression % (self.cleanValue(str(value)))

    # Clearing values from special characters.
    def CleanNode(self, node):
        search_ptrn = re.compile(r"[\/\\@?#&_%*',\(\)\" ]")
        replace_ptrn = re.compile(r"[ \/\\@?#&_%*',\(\)\" ]")
        match = search_ptrn.search(str(node))
        new_node = list()
        if match:
            replaced_str = replace_ptrn.sub('*', node)
            node = [x for x in replaced_str.split('*') if x]
            new_node.extend(node)
        else:
            new_node.append(node)
        node = new_node
        return node

    # Clearing values from special characters.
    def generateMapItemNode(self, node):
        key, value = node
        if key in self.allowedFieldsList:
            if self.mapListsSpecialHandling == False and type(value) in (
                    str, int, list) or self.mapListsSpecialHandling == True and type(value) in (str, int):
                return self.mapExpression % (key, self.generateCleanValueNodeLogsource(value))
            elif type(value) is list:
                return self.generateMapItemListNode(key, value)
            else:
                raise TypeError("Backend does not support map values of type " + str(type(value)))
        else:
            if self.mapListsSpecialHandling == False and type(value) in (
                    str, int, list) or self.mapListsSpecialHandling == True and type(value) in (str, int):
                if type(value) is str:
                    new_value = list()
                    value = self.CleanNode(value)
                    if type(value) == list:
                        new_value.append(self.andToken.join([self.valueExpression % val for val in value]))
                    else:
                        new_value.append(value)
                    if len(new_value)==1:
                        return "(" + self.generateANDNode(new_value) + ")"
                    else:
                        return "(" + self.generateORNode(new_value) + ")"
                else:
                    return self.generateValueNode(value)
            elif type(value) is list:
                new_value = list()
                for item in value:
                    item = self.CleanNode(item)
                    if type(item) is list and len(item) == 1:
                        new_value.append(self.valueExpression % item[0])
                    elif type(item) is list:
                        new_value.append(self.andToken.join([self.valueExpression % val for val in item]))
                    else:
                        new_value.append(item)
                return self.generateORNode(new_value)
            elif value is None:
                return self.nullExpression % (key, )
            else:
                raise TypeError("Backend does not support map values of type " + str(type(value)))

    # for keywords values with space
    def generateValueNode(self, node):
        if type(node) is int:
            return self.cleanValue(str(node))
        if 'AND' in node:
            return "(" + self.cleanValue(str(node)) + ")"
        else:
            return self.cleanValue(str(node))

    # collect elements of Arcsight search using OR
    def generateMapItemListNode(self, key, value):
        itemslist = list()
        for item in value:
            if key in self.allowedFieldsList:
                itemslist.append('%s = %s' % (key, self.generateValueNode(item)))
            else:
                itemslist.append('%s' % (self.generateValueNode(item)))
        return " OR ".join(itemslist)

    # prepare of tail for every translate
    def generate(self, sigmaparser):
        """Method is called for each sigma rule and receives the parsed rule (SigmaParser)"""
        const_title = ' AND type != 2 | rex field = flexString1 mode=sed "s//Sigma: {}/g"'
        for parsed in sigmaparser.condparsed:
            return self.generateQuery(parsed) + const_title.format(sigmaparser.parsedyaml["title"])

    # Add "( )" for values
    def generateSubexpressionNode(self, node):
        return self.subExpression % self.generateNode(node.items)

    # generateORNode algorithm for ArcSightBackend class.
    def generateORNode(self, node):
        if type(node) == ConditionOR and all(isinstance(item, str) for item in node):
            new_value = list()
            for value in node:
                value = self.CleanNode(value)
                if type(value) is list:
                    new_value.append(self.andToken.join([self.valueExpression % val for val in value]))
                else:
                    new_value.append(value)
            return "(" + self.orToken.join([self.generateNode(val) for val in new_value]) + ")"
        return "(" + self.orToken.join([self.generateNode(val) for val in node]) + ")"

class ArcSightESMBackend(SingleTextQueryBackend):
    """Converts Sigma rule into ArcSight ESM saved search. Contributed by SOC Prime. https://socprime.com"""
    reEscape = re.compile('(["\\\()])')
    identifier = "arcsight-esm"
    active = True
    andToken = " AND "
    orToken = " OR "
    notToken = " NOT "
    subExpression = "(%s)"
    listExpression = "(%s)"
    listSeparator = " OR "
    valueExpression = '"%s"'
    containsExpression = "%s CONTAINS %s"
    startsWithExpression = "%s CONTAINS %s"
    endsWithExpression = "%s CONTAINS %s"
    nullExpression = "NOT _exists_:%s"
    notNullExpression = "_exists_:%s"
    mapExpression = "%s = %s"
    mapListsSpecialHandling = True
    mapListValueExpression = "%s = %s"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        aFL = ["deviceVendor", "categoryDeviceGroup", "deviceProduct"]
        self.default_field = "deviceCustomString1"
        for item in self.sigmaconfig.fieldmappings.values():
            if item.target_type is list:
                aFL.extend(item.target)
            else:
                aFL.append(item.target)
        self.allowedFieldsList = list(set(aFL))

    # Skip logsource value from sigma document for separate path.
    def generateCleanValueNodeLogsource(self, value):
        return self.valueExpression % (self.cleanValue(str(value)))

    def CleanNode(self, node):
        if isinstance(node, str) and "*" in node and not node.startswith("*") and not node.endswith("*"):
            node = ["*{}*".format(x) for x in node.split('*') if x]
        return node

    #Clearing values from special characters.
    def generateMapItemNode(self, node):
        key, value = node
        value = self.CleanNode(value)
        if key in self.allowedFieldsList:
            if self.mapListsSpecialHandling == False and type(value) in (str, int, list) or self.mapListsSpecialHandling == True and type(value) in (str, int):
                if isinstance(value, str) and value.startswith("*") and value.endswith("*"):
                    return self.containsExpression % (
                    key, self.generateCleanValueNodeLogsource(value))
                elif isinstance(value, str) and value.startswith("*"):
                    return self.endsWithExpression % (key, self.generateCleanValueNodeLogsource(value))
                elif isinstance(value, str) and value.endswith("*"):
                    return self.startsWithExpression % (key, self.generateCleanValueNodeLogsource(value))
                else:
                    return self.mapExpression % (key, self.generateCleanValueNodeLogsource(value))
            elif type(value) is list:
                return self.generateMapItemListNode(key, value)
            else:
                raise TypeError("Backend does not support map values of type " + str(type(value)))
        else:
            key = self.default_field
            if self.mapListsSpecialHandling == False and type(value) in (str, int, list) or self.mapListsSpecialHandling == True and type(value) in (str, int):

                if isinstance(value, str) and value.startswith("*") and value.endswith("*"):
                    return self.containsExpression % (key, self.generateValueNode(self.CleanNode(value)))
                elif isinstance(value, str) and value.startswith("*"):
                    return self.endsWithExpression % (key, self.generateValueNode(self.CleanNode(value)))
                elif isinstance(value, str) and value.endswith("*"):
                    return self.startsWithExpression % (key, self.generateValueNode(self.CleanNode(value)))
                else:
                    return self.mapExpression % (key, self.generateValueNode(value))
            elif isinstance(value, list):
                new_value = list()
                for item in value:
                    item = self.CleanNode(item)
                    if type(item) is list and len(item) == 1:
                        new_value.append( self.containsExpression % (key, item[0]))
                    elif type(item) is list:
                        new_value.append(self.andToken.join([self.valueExpression % val for val in item]))
                    else:
                        if isinstance(item, str) and (item.startswith("*") or item.endswith("*")):
                            new_value.append(self.containsExpression % (key, self.generateValueNode(self.cleanValue(item))))
                        else:
                            new_value.append(self.mapExpression %(key, self.cleanValue(item)))
                return self.generateORNode(new_value)
            elif value is None:
                return self.nullExpression % (key, )
            else:
                raise TypeError("Backend does not support map values of type " + str(type(value)))

    # for keywords values with space
    def generateValueNode(self, node):
        if type(node) is int:
            return self.cleanValue(str(node))
        if 'AND' in node:
            return "(" + self.cleanValue(str(node)) + ")"
        else:
            return self.valueExpression % (self.cleanValue(str(node)))

    # collect elements of Arcsight search using OR
    def generateMapItemListNode(self, key, value):
        itemslist = list()
        for item in value:
            if isinstance(item, str) and (item.startswith("*") or item.endswith("*")):
                itemslist.append(self.containsExpression % (
                    key, self.generateValueNode(item)))
            else:
                itemslist.append(
                    '%s = %s' % (key, self.generateValueNode(item)))

        return " OR ".join(itemslist)

    # prepare of tail for every translate
    def generate(self, sigmaparser):
        """Method is called for each sigma rule and receives the parsed rule (SigmaParser)"""
        for parsed in sigmaparser.condparsed:
            return self.generateQuery(parsed)

    # Add "( )" for values
    def generateSubexpressionNode(self, node):
        return self.subExpression % self.generateNode(node.items)

    # generateORNode algorithm for ArcSightBackend class.
    def generateORNode(self, node):
        if type(node) == ConditionOR and all(isinstance(item, str) for item in node):
            new_value = list()
            for value in node:
                value = self.CleanNode(value)
                if type(value) is list:
                    new_value.append(self.andToken.join([self.valueExpression % val for val in value]))
                else:
                    new_value.append(value)
            return "(" + self.orToken.join([self.generateNode(val) for val in new_value]) + ")"
        elif isinstance(node, list) and all(isinstance(item, str) for item in node):
            return "(" + self.orToken.join([val for val in node]) + ")"
        else:
            return "(" + self.orToken.join([self.generateNode(val) for val in node]) + ")"



# Sigma value modifiers
# Copyright 2019 Thomas Patzke, Florian Roth

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import re
from .base import SigmaTransformModifier
from .mixins import ListOrStringModifierMixin
from sigma.parser.condition import ConditionAND, ConditionBase, ConditionOR, NodeSubexpression
from base64 import b64encode

class SigmaContainsModifier(ListOrStringModifierMixin, SigmaTransformModifier):
    """Add *-wildcard before and after all string(s)"""
    identifier = "contains"
    active = True
    valid_input_types = ListOrStringModifierMixin.valid_input_types + (NodeSubexpression,)

    def apply_str(self, val):
        try:
            if not val.startswith("*"):
                val = "*" + val
            if not val.endswith("*"):
                if val.endswith("\\"):
                    val += "\\*"
                else:
                    val += "*"
        except AttributeError:
            pass
        return val

class SigmaStartswithModifier(ListOrStringModifierMixin, SigmaTransformModifier):
    """Add *-wildcard before and after all string(s)"""
    identifier = "startswith"
    active = True

    def apply_str(self, val : str):
        if not val.endswith("*"):
            if val.endswith("\\"):
                val += "\\*"
            else:
                val += "*"
        return val

class SigmaEndswithModifier(ListOrStringModifierMixin, SigmaTransformModifier):
    """Add *-wildcard before and after all string(s)"""
    identifier = "endswith"
    active = True

    def apply_str(self, val : str):
        if not val.startswith("*"):
           val = '*' + val
        return val

class SigmaAllValuesModifier(SigmaTransformModifier):
    """Override default OR-linking behavior for list with AND-linking of all list values"""
    identifier = "all"
    active = True
    valid_input_types = (list, tuple, ConditionBase)

    def apply(self):
        vals = super().apply()
        cond = ConditionAND()
        for val in self.value:
            cond.add(val)
        return cond

    def validate(self):
        """Override default validate to avoid checking type recursively"""
        return any(( isinstance(self.value, t) for t in self.valid_input_types  ))

class SigmaBase64Modifier(ListOrStringModifierMixin, SigmaTransformModifier):
    """Encode strings with Base64"""
    identifier = "base64"
    active = True
    valid_input_types = ListOrStringModifierMixin.valid_input_types + (bytes,)

    def apply_str(self, val):
        if type(val) == str:
            val = val.encode()
        return b64encode(val).decode()

class SigmaBase64OffsetModifier(ListOrStringModifierMixin, SigmaTransformModifier):
    """Encode string(s) with Base64 in all three possible shifted offsets"""
    identifier = "base64offset"
    active = True
    valid_input_types = ListOrStringModifierMixin.valid_input_types + (bytes,)

    start_offsets = (0, 2, 3)
    end_offsets = (None, -3, -2)

    def apply_str(self, val):
        if type(val) == str:
            val = val.encode()
        items = [
                b64encode(
                    i * b' ' + val
                    )[
                        self.start_offsets[i]:
                        self.end_offsets[(len(val) + i) % 3]
                        ].decode()
                for i in range(3)
                ]
        cond = ConditionOR()
        cond.items = items
        return NodeSubexpression(cond)

class SigmaWindashModifier(ListOrStringModifierMixin, SigmaTransformModifier):
    """
    Expand parameter characters / and - that are often interchangeable in Windows into the other
    form if it appears between word boundaries. E.g. in -param-name the first dash will be expanded
    into /param-name while the second dash is left untouched.
    """
    identifier = "windash"
    active = True
    valid_input_types = ListOrStringModifierMixin.valid_input_types

    def expand_dashes(self, val, locations, offset=0):
        i = locations[0]
        if len(locations) == 1:
            subexpansions = [ val[i + 1:]]
        else:
            subexpansions = self.expand_dashes(val, locations[1:], i + 1)

        return [
            val[offset:i] + expanded + subexpansion
            for expanded in ("-", "/")
            for subexpansion in subexpansions
        ]

    def apply_str(self, val):
        dash_locations = [
            m.start()
            for m in re.finditer(re.compile("\\B[-/]\\b"), val)
        ]
        if dash_locations == []:
            return val
        else:
            items = self.expand_dashes(val, dash_locations)

        cond = ConditionOR()
        cond.items = items
        return NodeSubexpression(cond)

class SigmaEncodingBaseModifier(ListOrStringModifierMixin, SigmaTransformModifier):
    """
    Encode string to a byte sequence with the encoding given in the encoding property. This is
    a base class for all encoding modifiers.
    """
    identifier = "encoding-base"
    active = False
    encoding = "ascii"

    def apply_str(self, val):
        return val.encode(self.encoding)

class SigmaEncodeUTF16Modifier(SigmaEncodingBaseModifier):
    """Encode string to UTF-16 byte sequence"""
    identifier = "utf16"
    active = True
    encoding = "utf-16"

class SigmaEncodeUTF16LEModifier(SigmaEncodingBaseModifier):
    """Encode string to UTF-16 little endian byte sequence"""
    identifier = "utf16le"
    active = True
    encoding = "utf-16le"

class SigmaEncodeWideModifier(SigmaEncodeUTF16LEModifier):
    """Modifier 'wide' is an alias for the utf16le modifier."""
    identifier = "wide"

class SigmaEncodeUTF16BEModifier(SigmaEncodingBaseModifier):
    """Encode string to UTF-16 big endian byte sequence"""
    identifier = "utf16be"
    active = True
    encoding = "utf-16be"



# Output backends for sigmac
# Copyright 2016-2018 Thomas Patzke, Florian Roth (Nextron Systems), Roey, Karneades

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import re
import sigma
from .base import SingleTextQueryBackend
from .mixins import MultiRuleOutputMixin

class PowerShellBackend(SingleTextQueryBackend):
    """Converts Sigma rule into PowerShell event log cmdlets."""
    identifier = "powershell"
    active = True
    config_required = False
    default_config = ["sysmon", "powershell"]
    options = (
        ("csv", False, "Return the results in CSV format instead of Powershell objects", None),
    )

    reEscape = re.compile('("|(?<!\\\\)\\\\(?![*?\\\\])|\+)')
    reClear = None
    andToken = " -and "
    orToken = " -or "
    notToken = " -not "
    subExpression = "(%s)"
    listExpression = "($_.message -match %s)"
    listSeparator = "$_.message -match "
    valueExpression = "\"%s\""
    nullExpression = "-not %s=\"*\""
    notNullExpression = "%s=\"*\""
    mapExpression = "$_.%s -eq %s"
    mapListsSpecialHandling = True

    logname = None
    fieldMappings = {
        "EventID": "ID",
        "ID": "ID",
        "ServiceFileName": "Service File Name"
    }

    def generate(self, sigmaparser):
        """Method is called for each sigma rule and receives the parsed rule (SigmaParser)"""
        for parsed in sigmaparser.condparsed:
            query = self.generateQuery(parsed, sigmaparser)
            before = self.generateBefore(parsed)
            after = self.generateAfter(parsed)

            result = ""

            if before is not None:
                result = before
            if query is not None:
                result += query
            if after is not None:
                result += after

            return result

    def generateBefore(self, parsed):
        if self.logname:
            return "Get-WinEvent -LogName %s | where {" % self.logname
        return "Get-WinEvent | where {"

    def generateAfter(self, parsed):
        if self.csv:
            return " | ConvertTo-CSV -NoTypeInformation"
        return ""

    def generateNode(self, node):
        if type(node) == sigma.parser.condition.ConditionAND:
            return self.generateANDNode(node)
        elif type(node) == sigma.parser.condition.ConditionOR:
            return self.generateORNode(node)
        elif type(node) == sigma.parser.condition.ConditionNOT:
            return self.generateNOTNode(node)
        elif type(node) == sigma.parser.condition.ConditionNULLValue:
            return self.generateNULLValueNode(node)
        elif type(node) == sigma.parser.condition.ConditionNotNULLValue:
            return self.generateNotNULLValueNode(node)
        elif type(node) == sigma.parser.condition.NodeSubexpression:
            return self.generateSubexpressionNode(node)
        elif type(node) == tuple:
            return self.generateMapItemNode(node)
        elif type(node) in (str, int):
            return self.generateValueNode(node, False)
        elif type(node) == list:
            return self.generateListNode(node)
        else:
            raise TypeError("Node type %s was not expected in Sigma parse tree" % (str(type(node))))

    def generateQuery(self, parsed, sigmaparser):
        result = self.generateNode(parsed.parsedSearch)
        self.parsedlogsource = sigmaparser.get_logsource().service

        powershellPrefix = ""
        if parsed.parsedAgg:
            powershellSuffixAgg = self.generateAggregation(parsed.parsedAgg)
            result = result + " } " + powershellSuffixAgg
        else:
            result = powershellPrefix + result + " } | select TimeCreated,Id,RecordId,ProcessId,MachineName,Message"
        return result

    def generateMapItemNode(self, node):
        key, value = node
        if self.mapListsSpecialHandling == False and type(value) in (str, int, list) or self.mapListsSpecialHandling == True and type(value) in (str, int):
            if key in ("LogName","source"):
                self.logname = value
            elif key in self.fieldMappings.keys():
                key = self.fieldMappings[key]
                return self.mapExpression % (key, self.generateValueNode(value, True))
            elif type(value) == str and "*" in value:
                value = value.replace("*", ".*")
                if key == "Message":
                    return "$_.message -match %s" % (self.generateValueNode(value, True))
                else:
                    return "$_.message -match %s" % (self.generateValueNode(key + ".*" + value, True))
            elif type(value) in (str, int):
                return '$_.message -match %s' % (self.generateValueNode(key + ".*" +str(value), True))
            else:
                return self.mapExpression % (key, self.generateNode(value))
        elif type(value) == list:
            return self.generateMapItemListNode(key, value)
        elif value is None:
            return self.nullExpression % (key, )
        else:
            raise TypeError("Backend does not support map values of type " + str(type(value)))

    def generateMapItemListNode(self, key, value):
        itemslist = list()
        for item in value:
            if key in self.fieldMappings.keys():
                key = self.fieldMappings[key]
                itemslist.append(self.mapExpression % (key, self.generateValueNode(item, True)))
            elif type(item) == str and "*" in item:
                item = item.replace("*", ".*")
                if key == "Message":
                    itemslist.append('$_.message -match %s' % (self.generateValueNode(item, True)))
                else:
                    itemslist.append('$_.message -match %s' % (self.generateValueNode(key + ".*" +item, True)))
            else:
                itemslist.append('$_.message -match %s' % (self.generateValueNode(item, True)))
        return '('+" -or ".join(itemslist)+')'

    def generateANDNode(self, node):
        generated = [ self.generateNode(val) for val in node ]
        filtered = [ g for g in generated if g is not None ]
        if filtered:
            return self.andToken.join(filtered)
        else:
            return None

    def generateValueNode(self, node, keypresent):
        if keypresent == False:
            return "$_.message -match \"{0}\"".format(str(node))
        else:
            return self.valueExpression % (self.cleanValue(str(node)))

    def getPowerShellCondOp(self, cond_op):
        if(cond_op == "<"):
            return "-lt"
        elif(cond_op == ">"):
            return "-gt"
        elif(cond_op == "="):
            return "-eq"

    def generateAggregation(self, agg):
        if agg == None:
            return ""
        if agg.aggfunc != sigma.parser.condition.SigmaAggregationParser.AGGFUNC_COUNT:
            raise NotImplementedError("Only COUNT aggregation function is implemented for this backend")
        if agg.aggfunc == sigma.parser.condition.SigmaAggregationParser.AGGFUNC_NEAR:
            # python .\tools\sigmac -t splunk -c .\tools\config\splunk-windows-all.yml -r .\rules\windows\builtin\
            # Example rule: .\sigma\rules\windows\builtin\win_susp_samr_pwset.yml
            raise NotImplementedError("The 'near' aggregation operator is not yet implemented for this backend")
        if agg.groupfield == None:
            # Example rule: .\sigma\rules\windows\builtin\win_multiple_suspicious_cli.yml
            powershell_cond_op = self.getPowerShellCondOp(agg.cond_op)
            return " | group-object %s | where { $_.count %s %s } | select name,count | sort -desc" % (agg.aggfield or "", powershell_cond_op, agg.condition)
        else:
            # Example rule: .\sigma\rules\windows\other\win_rare_schtask_creation.yml
            powershell_cond_op = self.getPowerShellCondOp(agg.cond_op)
            if (agg.aggfield == None):
                return " | group-object %s | where { $_.count %s %s } | select name,count | sort -desc" % (agg.groupfield or "", powershell_cond_op, agg.condition)
            else:
                return " | select %s, %s | group %s | foreach { [PSCustomObject]@{'%s'=$_.name;'Count'=($_.group.%s | sort -u).count} }  | sort count -desc | where { $_.count %s %s }" % (agg.groupfield, agg.aggfield, agg.groupfield, agg.groupfield, agg.aggfield, powershell_cond_op, agg.condition)



# Output backends for sigmac
# Copyright 2016-2018 Thomas Patzke, Florian Roth (Nextron Systems), Devin Ferguson, Julien Bachmann

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import json
import re
import sys
import os
from typing import List, Tuple, Union
from uuid import uuid4

from sigma.parser.condition import SigmaAggregationParser

from .elasticsearch import ElasticsearchQuerystringBackend

class Atom:
    def __init__(self, field: str, prop: str) -> None:
        self.field = field
        self.prop = prop
    def __str__(self) -> str:
        return "Atom( {}, {} )".format(self.field.replace("\\\\", "\\"), self.prop.replace("\\\\", "\\"))

# Root of AST is always a Group
class Group:
    def __init__(self) -> None:
        pass
    def __str__(self) -> str:
        return "Group( {} )".format(str(self.ary).replace("\\\\", "\\"))

class Boolean:
    def __init__(self, expression: Union[Atom, Group]) -> None:
        self.expression = expression
    def __str__(self) -> str:
        return "Boolean( {} )".format(str(self.expression).replace("\\\\", "\\"))

class Ary:
    def __init__(self, bool1: Boolean, bool2: List[Tuple[str, Boolean]] = None) -> None:
        self.bool1 = bool1
        self.bool2 = bool2
    def __str__(self) -> str:
        return "Ary( {}, {} )".format(str(self.bool1).replace("\\\\", "\\"), [(rel, str(boolean).replace("\\\\", "\\")) for rel, boolean in self.bool2])

def group_init(self, ary: Ary):
    self.ary = ary

Group.__init__ = group_init

def parse_atom(s: str) -> Atom:
    reg = r"(?<!\\):" # (any character that's not '\') followed by ':'
    return Atom(*re.split(reg, s))

'''
Since root of AST is always a Group, call parse_group to initiate parsing of overall expression.
'''
def parse_group(s: str) -> Group:
    return Group(parse_ary(s[1:-1]))

'''
Expand special group in form of A:(B OR C) to (A:B OR A:C)
'''
def expand_group(s: str) -> str:
    reg = r"(?<!\\):" # (any character that's not '\') followed by ':'

    field, props = re.split(reg, s.strip("()")) # props = (prop1 OR prop2...)
    props = props.strip("()").split() # Further split props
    newGroup = []

    for index in range(len(props)):
        element = props[index]
        if element not in ["AND", "OR"]:
            newGroup.append(f'{field}:{element}')
        else:
            newGroup.append(element)

    return "(" + " ".join(newGroup) + ")"

def parse_boolean(s: str) -> Boolean:
    if "(" not in s:
        expression = parse_atom(s)
    else:
        if s[0] != '(':
            s = expand_group(s)
        expression = parse_group(s)

    return Boolean(expression)

def parse_ary(s: str) -> Ary:
    lst = []
    left = right = level = 0

    while left < len(s):
        # Going down one level
        if right < len(s) and s[right] == '(':
            level += 1

        # Going up one level
        elif right < len(s) and s[right] == ')':
            level -= 1

        # s[left:right] is parse-able
        elif right == len(s) or (s[right] == ' ' and level == 0):
            section = s[left:right]

            # Handle Boolean case
            if section not in ["AND", "OR"]:
                section = parse_boolean(section)

            lst.append(section)
            left = right + 1

        right += 1

    # [Bool, Rel, Bool, Rel, Bool,...] => Bool, [(Rel, Bool), (Rel, Bool),...]
    bool1 = lst[0]
    bool2 = []

    for i in range(1, len(lst), 2):
        tupe = (lst[i], lst[i + 1])
        bool2.append(tupe)
    
    return Ary(bool1, bool2)

def translate_atom(atom: Atom) -> dict:
    return {
        "match": {
            atom.field: atom.prop
        }
    }

def translate_group(group: Group) -> dict:
    return translate_ary(group.ary)

def translate_boolean(boolean: Boolean) -> dict:
    if type(boolean.expression) is Atom:
        return translate_atom(boolean.expression)

    return translate_group(boolean.expression)

'''
Combining ary.bool1 and ary.bool2 into array of Boolean grouped by ANDs and split by ORs.
''' 
def convert_bool_array(bool1: Boolean, boolArr: List[Tuple[str, Boolean]]) -> List[List[Boolean]]:
    result = [[bool1]]
    resultIndex = 0
    
    for rel, boolean in boolArr:
        if rel == "AND":
            if resultIndex == len(result):
                result.append([boolean])
            else:
                result[resultIndex].append(boolean)
        else:
            resultIndex += 2
            result.append([boolean])

    return result

'''
Group atomic match statements together into parent clause and wrap inside bool statement.
Maintain group match statements, which are already wrapped in bool statement.
'''
def adjust_matches(matches: List[dict], clause) -> List[dict]:
    atomicMatches = []
    combinedAtomicMatches = []
    groupMatches = []

    # Determine if current statement is an atomic match or bool group statement
    for index in range(len(matches)):
        match = matches[index]
        if "match" in match.keys():
            atomicMatches.append(match)
        else:
            groupMatches.append(match)

    # If any atomic matches, combine under parent clause wrapped in a single bool statement
    if atomicMatches:
        # If there's only one atomic match, it should be wrapped in a bool-must regardless of the parent clause
        clause = "must" if len(atomicMatches) == 1 else clause

        combinedAtomicMatches = [{
                "bool": {
                    clause: atomicMatches
                }
            }]
    
    return combinedAtomicMatches + groupMatches

def contains_group(booleanArr: List[Boolean]) -> bool:
    for boolean in booleanArr:
        if type(boolean.expression) is Group:
            return True
    
    return False

def translate_ary(ary: Ary) -> dict:
    parsedTranslation = convert_bool_array(ary.bool1, ary.bool2)

    clauses = []
    
    translateIndex = 0
    while translateIndex < len(parsedTranslation):
        parsedExpression = parsedTranslation[translateIndex]
        currMatches = []
        clause = "must" # default clause is "must"; clause is "should" if multiple consecutive "or" statements

        # Statement was joined by "or"
        if len(parsedExpression) == 1:
            counter = 1
            tempIndex = translateIndex
            while tempIndex+1 < len(parsedTranslation) and len(parsedTranslation[tempIndex+1]) == 1:
                tempIndex += 1
                counter += 1

            # If there's more than one, use "should" clause instead of "must"
            if counter > 1:
                clause = "should"
                parsedExpression = []

                # Rebuild parsed expression to join statements together and fast forward the translate index
                for i in range(counter):
                    parsedExpression += parsedTranslation[translateIndex+i]
                
                translateIndex = tempIndex
        
        # Iterate through each statement and join match statements into array
        for boolean in parsedExpression:
            currMatches.append(translate_boolean(boolean))
        
        # If bool array contains a Group which is wrapped in a bool, match statements must also be wrapped in a bool.
        if contains_group(parsedExpression):
            currMatches = adjust_matches(currMatches, clause)

        currQuery = {
            "bool": {
                clause: currMatches
            }
        }

        clauses.append(currQuery)
        translateIndex += 1

    # If only one type of clause, don't use nested bool object
    if len(clauses) > 1:
        return {
                        "bool": {
                            "should": clauses
                        }
                    }
    return clauses[0]

class OpenSearchBackend(object):
    """OpenSearch detection rule backend."""
    active = True
    uuid_black_list = []
    options = ElasticsearchQuerystringBackend.options + (
                ("put_filename_in_ref", False, "Want to have yml name in reference ?", None),
                ("convert_to_url", False, "Want to convert to a URL ?", None),
                ("path_to_replace", "../", "The local path to replace with dest_base_url", None),
                ("dest_base_url", "https://github.com/SigmaHQ/sigma/tree/master/", "The URL prefix", None),
                ("custom_tag", None , "Add custom tag. for multi split with a comma tag1,tag2 ", None),
            )
    isThreshold = False

    # Default values for fields exclusive to OpenSearch monitors
    RULE_TYPE = "monitor"
    IS_ENABLED = True
    INTERVAL = 5
    UNIT = "MINUTES"
    TRIGGER_NAME = "generated-trigger"
    SEVERITIES = {"informational": "5", "low": "4", "medium": "3", "high": "2", "critical": "1"}
    TRIGGER_SCRIPT = "ctx.results[0].hits.total.value > 0"
    TRIGGER_LANGUAGE = "painless"
    MONITOR_INDICES = ["opensearch-security-logs"]
    NUM_RESULTS = 1

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.tactics = self._load_mitre_file("tactics")
        self.techniques = self._load_mitre_file("techniques")
        self.rule_threshold = {}

    '''
    Loads appropriate mitre file and returns mappings as dict.
    '''
    def _load_mitre_file(self, mitre_type):
        try:
            backend_dir = os.path.normpath(os.path.join(os.path.dirname(os.path.abspath(__file__)), "..", "..", "config", "mitre"))
            path = os.path.join(backend_dir,"{}.json".format(mitre_type))
            with open(path, 'r') as config_file:
                config = json.load(config_file)
                return config
        except (IOError, OSError) as e:
            print("Failed to open {} configuration file '%s': %s".format(path, str(e)), file=sys.stderr)
            return []
        except json.JSONDecodeError as e:
            print("Failed to parse {} configuration file '%s' as valid YAML: %s" % (path, str(e)), file=sys.stderr)
            return []

    '''
    Calls parent generate methods to retrieve Sigma rule condition as Elastic Common Schema query.
    Then calls the create_rule method to return final translated object.
    '''
    def generate(self, sigmaparser):
        # reset per-detection variables
        self.rule_threshold = {}
        translation = super().generate(sigmaparser)
        if translation:
            index = sigmaparser.get_logsource().index
            if len(index) == 0:
                index = ["apm-*-transaction", "auditbeat-*", "endgame-*", "filebeat-*", "packetbeat-*", "winlogbeat-*"]
            configs = sigmaparser.parsedyaml
            configs.update({"translation": translation})
            rule = self.create_rule(configs, index)
            return rule

    '''
    Generates threat detection for OpenSearch monitor, which compiles tactics and techniques found in Sigma tags.
    '''
    def create_threat_description(self, tactics_list, techniques_list):
        threat_list = list()
        for tactic in tactics_list:
            temp_tactics = {
                "tactic": {
                    "id": tactic.get("external_id", ""),
                    "reference": tactic.get("url", ""),
                    "name": tactic.get("tactic", "")
                },
                "framework": "MITRE ATT&CK®"
            }
            temp_techniques = list()
            for tech in techniques_list:
                if tactic.get("tactic", "") in tech.get("tactic", []):
                    temp_techniques.append({
                                "id": tech.get("technique_id", ""),
                                "name": tech.get("technique", ""),
                                "reference": tech.get("url", "")
                            })
            temp_tactics.update({"technique": temp_techniques})
            threat_list.append(temp_tactics)
        return threat_list

    '''
    Finds tactics mentioned in Sigma tags.
    '''
    def find_tactics(self, key_name=None, key_id=None):
        for tactic in self.tactics:
            if key_name and key_name == tactic.get("tactic", ""):
                return tactic
            if key_id and key_id == tactic.get("external_id", ""):
                return tactic

    '''
    Finds techniques mentioned in Sigma tags.
    '''
    def find_technique(self, key_id=None):
        for technique in self.techniques:
            if key_id and key_id == technique.get("technique_id", ""):
                return technique

    '''
    Maps Sigma severity to OpenSearch numerical severity from 1-5.
    '''
    def map_severity(self, severity):
        severity = severity.lower()
        return self.SEVERITIES[severity] if severity in self.SEVERITIES else self.SEVERITIES["medium"]

    def create_trigger(self, severity):
        return [
                {
                "name": self.TRIGGER_NAME,
                "severity": self.map_severity(severity),
                "condition": {
                    "script": {
                        "source": f'{self.TRIGGER_SCRIPT}',
                        "lang": self.TRIGGER_LANGUAGE
                    }
                },
                "actions": []
                }
            ]

    def build_threshold(self, field, inequality, threshold):
        INEQUALITIES = {"<": "lt", "<=": "lte", ">": "gt", ">=": "gte"}

        return {
            "range": {
                field: {
                    INEQUALITIES[inequality]: threshold
                }
            }
        }

    '''
    Builds OpenSearch monitor query from translated Elastic Rule query. Forms an abstract syntax tree (AST)
    using the following repeated structures:
    - Atom = A:B
    - Rel = AND | OR
    - Ary = Bool [Rel Bool]*
    - Group = (Ary)
    - SGroup = A:(B OR C)
    - Bool = Atom | Group | SGroup

    Then translates AST into OpenSearch boolean queries.
    '''
    def build_query(self, translation):
        ast = parse_group(translation)
        translatedQuery = translate_group(ast)

        if self.isThreshold:
            translatedQuery["bool"]["filter"] = self.rule_threshold
        
        return translatedQuery

    '''
    Builds inputs field of OS monitor.
    '''
    def build_inputs(self, translation):
        return [
                {
                    "search": {
                        "indices": self.MONITOR_INDICES,
                        "query": {
                            "size": self.NUM_RESULTS,
                            "aggregations": {},
                            "query": self.build_query(translation)
                        }
                    }
                }
            ]

    '''
    Adds Sigma yml file name in references if self.put_filename_in_ref option is True.
    '''
    def build_ymlfile_ref(self, configs):
        if self.put_filename_in_ref == False:  # Dont want
            return None
        yml_filename = configs.get("yml_filename")
        yml_path = configs.get("yml_path")
        if yml_filename == None or yml_path == None:
            return None
            
        if self.convert_to_url:
            yml_path = yml_path.replace('\\','/')                              #windows path to url 
            self.path_to_replace = self.path_to_replace.replace('\\','/')      #windows path to url            
            if self.path_to_replace not in yml_path: #Error to change
                return None

            new_ref = yml_path.replace(self.path_to_replace,self.dest_base_url) + '/' + yml_filename
        else:
            new_ref = yml_filename
        return new_ref

    '''
    Builds the list of searchable tags. Matches against list of known tags and adds any custom tags.
    '''
    def build_tags_list(self, tags):
        tactics_list = list()
        new_tags = list()
        technics_list = list()

        for tag in tags:
            tag = tag.replace("attack.", "")
            if re.match("[t][0-9]{4}", tag, re.IGNORECASE):
                tech = self.find_technique(tag.title())
                if tech:
                    new_tags.append(tag.title())
                    technics_list.append(tech)
            else:
                if "_" in tag:
                    tag_list = tag.split("_")
                    tag_list = [item.title() for item in tag_list]
                    tact = self.find_tactics(key_name=" ".join(tag_list))
                    if tact:
                        new_tags.append(" ".join(tag_list))
                        tactics_list.append(tact)
                elif re.match("[ta][0-9]{4}", tag, re.IGNORECASE):
                    tact = self.find_tactics(key_id=tag.upper())
                    if tact:
                        new_tags.append(tag.upper())
                        tactics_list.append(tact)
                else:
                    tact = self.find_tactics(key_name=tag.title())
                    if tact:
                        new_tags.append(tag.title())
                        tactics_list.append(tact)

        if self.custom_tag:
            if ',' in self.custom_tag:
                tag_split = self.custom_tag.split(",")
                for l_tag in tag_split:
                    new_tags.append(l_tag)   
            else:    
                new_tags.append(self.custom_tag)
        
        return tactics_list, technics_list, new_tags

    '''
    Get the rule id of the Sigma rule. If the rule id is blank or isn't unique, generate a random one.
    '''
    def get_rule_id(self, rule_uuid):
        rule_uuid = rule_uuid.lower()
        if rule_uuid == "" or rule_uuid in self.uuid_black_list:
            rule_uuid = str(uuid4())
        self.uuid_black_list.append(rule_uuid)
        rule_id = re.sub(re.compile('[()*+!,\[\].\s"]'), "_", rule_uuid)

        return rule_id

    '''
    Gets list of references.
    '''
    def get_references(self, configs):
        references = configs.get("reference") if configs.get("reference") is not None else configs.get("references")
        references = self.build_ref_yaml(references, configs)
        return references

    '''
    Adds Sigma yml file to references.
    '''
    def build_ref_yaml(self, references, configs):
        add_ref_yml = self.build_ymlfile_ref(configs)
        if add_ref_yml:
            if references is None: # No ref
                references=[]
            if add_ref_yml in references:
                pass # else put a duplicate ref for  multi rule file
            else:
                references.append(add_ref_yml)
        
        return references

    '''
    Main method that builds OpenSearch monitor and returns it in JSON format.
    '''
    def create_rule(self, configs, index):
        rule_name = configs.get("title", "")
        
        rule_description = configs.get("description", "")
        
        inputs = self.build_inputs(configs.get("translation", ""))
        
        triggers = self.create_trigger(configs.get("level", "medium"))
        
        rule_id = self.get_rule_id(configs.get("id", ""))

        tactics_list, technics_list, new_tags = self.build_tags_list(configs.get("tags", []))   
        threat = self.create_threat_description(tactics_list, technics_list)
        
        references = self.get_references(configs)
        
        rule = {
            "type": self.RULE_TYPE,
            "name": rule_name,
            "description": rule_description,
            "enabled": self.IS_ENABLED,
            "schedule": {
                "period": {
                    "interval": self.INTERVAL,
                    "unit": self.UNIT
                }
            },
            "inputs": inputs,
            "tags": new_tags,
            "triggers": triggers,
            "sigma_meta_data": {
                "rule_id": rule_id,
                "threat": threat
            }
        }

        if references:
            rule.update({"references": references})

        return json.dumps(rule)

class OpenSearchQsBackend(OpenSearchBackend, ElasticsearchQuerystringBackend):
    '''
    Backend class containing the identifier for the -t argument. Can inherit from ElasticsearchQuerystringBackend
    since query string in both OpenSearch monitors and ElasticRule are in Elastic Common Schema.
    '''
    identifier = "opensearch-monitor"
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def generateAggregation(self, agg):
        if agg.aggfunc == SigmaAggregationParser.AGGFUNC_COUNT:
            if agg.aggfield:
                raise NotImplementedError("Threshold rules cannot COUNT(DISTINCT %s)" % agg.aggfield)
            self.isThreshold = True
            self.rule_threshold = self.build_threshold(agg.groupfield, agg.cond_op, agg.condition)
            return ""
        raise NotImplementedError("Aggregation %s is not implemented for this backend" % agg.aggfunc_notrans)


# Default Datamodel Mappings for SplunkDMBackend

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

default_datamodels = {
    "Endpoint": {
        "datasets": {
            "Processes": {
                "fields": {
                    "action": ["EventType"],
                    "cpu_load_percent ": [],
                    "dest": ["ComputerName", "Computer"],
                    "dest_bunit": [],
                    "dest_category": [],
                    "dest_is_expected": [],
                    "dest_priority": [],
                    "dest_requires_av": [],
                    "dest_should_timesync": [],
                    "dest_should_update": [],
                    "mem_used": [],
                    "original_file_name": ["OriginalFileName"],
                    "os": [],
                    "parent_process": ["ParentCommandLine"],
                    "parent_process_exec": [],
                    "parent_process_id": ["ParentProcessId"],
                    "parent_process_guid": ["ParentProcessGuid"],
                    "parent_process_name": [""],
                    "parent_process_path": ["ParentImage"],
                    "process": ["CommandLine"],
                    "process_current_directory": [],
                    "process_exec": [],
                    "process_hash": ["Hashes"],
                    "process_id": ["ProcessId"],
                    "process_guid": ["ProcessGuid"],
                    "process_integrity_level": ["IntegrityLevel"],
                    "process_name": [],
                    "process_path": ["Image"],
                    "tag": [],
                    "user": ["User", "SubjectUserName", "AccountName"],
                    "user_id": ["LogonGuid"],
                    "user_bunit": [],
                    "user_category": [],
                    "user_priority": [],
                    "vendor_product": []
                },
                "mapping": {
                    "category": "process_creation"
                }
            },
            "Filesystem": {
                "fields": {
                    "action": ["EventType"],
                    "dest": ["ComputerName", "Computer"],
                    "dest_bunit": [],
                    "dest_category": [],
                    "dest_priority": [],
                    "dest_requires_av": [],
                    "dest_should_timesync": [],
                    "dest_should_update": [],
                    "file_access_time": [],
                    "file_create_time": ["CreationUtcTime"],
                    "file_hash": ["Hash"],
                    "file_modify_time": [],
                    "file_name": [],
                    "file_path": ["TargetFilename"],
                    "file_acl": [],
                    "file_size": [],
                    "process_id": ["ProcessId"],
                    "process_guid": ["ProcessGuid"],
                    "tag": [],
                    "user": ["User", "SubjectUserName", "AccountName"],
                    "user_bunit": [],
                    "user_category": [],
                    "user_priority": [],
                    "vendor_product": []
                },
                "mapping": {
                    "category": "file_event"
                }
            },
            "Registry": {
                "fields": {
                    "action": ["EventType"],
                    "dest": ["ComputerName", "Computer"],
                    "dest_bunit": [],
                    "dest_category": [],
                    "dest_priority": [],
                    "dest_requires_av": [],
                    "dest_should_timesync": [],
                    "dest_should_update": [],
                    "process_id": ["ProcessId"],
                    "process_guid": ["ProcessGuid"],
                    "registry_hive": [],
                    "registry_path": ["TargetObject"],
                    "registry_key_name": [],
                    "registry_value_data": ["Details"],
                    "registry_value_name": [],
                    "registry_value_text": [],
                    "registry_value_type": [],
                    "status": [],
                    "tag": [],
                    "user": ["User", "SubjectUserName", "AccountName"],
                    "user_bunit": [],
                    "user_category": [],
                    "user_priority": [],
                    "vendor_product": []
                },
                "mapping": {
                    "category": "registry_event",
                    "category": "registry_add",
                    "category": "registry_delete",
                    "category": "registry_set",
                    "category": "registry_rename"
                }
            }
        }
    }
}



# Sigma parser
# Copyright 2016-2018 Thomas Patzke, Florian Roth

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import yaml
from sigma.parser.condition import ConditionAND, ConditionOR
from sigma.config.exceptions import SigmaConfigParseError
from sigma.config.mapping import FieldMapping, FieldMappingChain

# Chain of multiple configurations
class SigmaConfigurationChain(list):
    """
    Chain of SigmaConfiguration objects. Behaves like a list of Sigma configuration objects on the one side and
    like a SigmaConfiguration object on the other. All methods are applied to the given parameters in the order
    of addition of the configurations.
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.backend = None
        self.defaultindex = None
        self.config = dict()
        self.fieldmappings = dict()
        self.logsources = dict()

        for config in self:
            self.postprocess_config(config)

    def append(self, config):
        super().append(config)
        self.postprocess_config(config)

    def postprocess_config(self, config):
        self.defaultindex = config.defaultindex
        self.config.update(config.config)
        self.fieldmappings.update(config.fieldmappings)
        self.logsources.update(config.logsources)

    def get_fieldmapping(self, fieldname):
        """Return mapped fieldname by iterative application of each config stored in configuration chain."""
        if self:
            fieldmappings = FieldMappingChain(fieldname)
            for config in self:
                fieldmappings.append(config)
            return fieldmappings
        else:
            return FieldMapping(fieldname)

    def get_logsource(self, category, product, service):
        """Return merged log source definition of all logosurces that match criteria across all Sigma conversion configurations in chain."""
        matching = list()
        for config in self:
            for logsource in config.logsources:
                if logsource.matches(category, product, service):
                    matching.append(logsource)
                    if logsource.rewrite is not None:
                        category, product, service = logsource.rewrite
        return SigmaLogsourceConfiguration(matching, self.defaultindex)

    def get_logsourcemerging(self):
        value = ''
        for config in self:
            if value == '':
                value =  config.get_logsourcemerging().lower()

        if not value in ['and', 'or']:
            value = 'and'

        return value

    def set_backend(self, backend):
        """Set backend for all sigma conversion configurations in chain."""
        self.backend = backend
        for config in self:
            config.set_backend(backend)

    def get_indexfield(self):
        """Get index condition if index field name is configured"""
        if self.backend is not None:
            return self.backend.index_field

# Configuration
class SigmaConfiguration:
    """Sigma converter configuration. Contains field mappings and logsource descriptions"""
    def __init__(self, configyaml=None):
        if configyaml == None:
            self.config = None
            self.order = None
            self.fieldmappings = dict()
            self.logsources = dict()
            self.defaultindex = None
            self.backend = None
        else:
            config = yaml.safe_load(configyaml)
            self.config = config

            self.fieldmappings = dict()
            try:
                for source, target in config['fieldmappings'].items():
                    self.fieldmappings[source] = FieldMapping(source, target)
            except TypeError as e:
                raise SigmaConfigParseError("Configuration has wrong type, should be map") from e
            except KeyError:
                pass

            if type(self.fieldmappings) != dict:
                raise SigmaConfigParseError("Fieldmappings must be a map")

            self.order = config.setdefault("order", None)
            self.defaultindex = config.setdefault('defaultindex', None)

            self.logsources = list()
            self.backend = None

    def get_fieldmapping(self, fieldname):
        """Return mapped fieldname if mapping defined or field name given in parameter value"""
        try:
            return self.fieldmappings[fieldname]
        except KeyError:
            return FieldMapping(fieldname)

    def get_logsource(self, category, product, service):
        """Return merged log source definition of all logosurces that match criteria"""
        matching = [logsource for logsource in self.logsources if logsource.matches(category, product, service)]
        return SigmaLogsourceConfiguration(matching, self.defaultindex)

    def get_logsourcemerging(self):
        if self.config != None:
            if 'logsourcemerging' in self.config:
                return self.config['logsourcemerging']
        return ''

    def set_backend(self, backend):
        """Set backend. This is used by other code to determine target properties for index addressing"""
        self.backend = backend
        if self.config != None:
            if 'logsources' in self.config:
                logsources = self.config['logsources']
                if type(logsources) != dict:
                    raise SigmaConfigParseError("Logsources must be a map")
                for name, logsource in logsources.items():
                    self.logsources.append(SigmaLogsourceConfiguration(logsource, self.defaultindex))

    def get_indexfield(self):
        """Get index condition if index field name is configured"""
        if self.backend is not None:
            return self.backend.index_field

class SigmaLogsourceConfiguration:
    """Contains the definition of a log source"""
    def __init__(self, logsource=None, defaultindex=None):
        self.search = []
        if logsource == None:               # create empty object
            self.merged = False
            self.category = None
            self.product = None
            self.service = None
            self.index = list()
            self.conditions = list()    # a list of (field, value) tuples which are OR-linked in the generated query. May also contain such a list as list element (in case of merged log sources)
            self.rewrite = None
        elif type(logsource) == list and all([isinstance(o, SigmaLogsourceConfiguration) for o in logsource]):      # list of SigmaLogsourceConfigurations: merge
            self.merged = True
            if any([ ls.merged for ls in logsource ]):      # Ensure that already merged objects are not merged again
                raise TypeError("Nested merging of SigmaLogsourceConfiguration objects is not allowed")
            rewrites = { ls.rewrite for ls in logsource if ls.rewrite is not None }
            if len(rewrites) > 1:
                raise ValueError("More than one matching log source contains a rewrite part")
            elif len(rewrites) == 1:
                self.rewrite = rewrites.pop()
            else:
                self.rewrite = None

            # Merge category, product and service
            categories = { ls.category for ls in logsource if ls.category is not None }
            products = { ls.product for ls in logsource if ls.product is not None }
            services = { ls.service for ls in logsource if ls.service is not None }
            if len(categories) > 1 or len(products) > 1 or len(services) > 1:
                raise ValueError("Merged SigmaLogsourceConfigurations must have disjunct categories (%s), products (%s) and services (%s)" % (str(categories), str(products), str(services)))

            try:
                self.category = categories.pop()
            except KeyError:
                self.category = None
            try:
                self.product = products.pop()
            except KeyError:
                self.product = None
            try:
                self.service = services.pop()
            except KeyError:
                self.service = None

            # Merge all index patterns
            self.index = list(set([index for ls in logsource for index in ls.index]))       # unique(flat(logsources.index))
            if len(self.index) == 0 and defaultindex is not None:   # if no index pattern matched and default index is present: use default index
                if type(defaultindex) == str:
                    self.index = [defaultindex]
                elif type(defaultindex) == list and all([type(i) == str for i in defaultindex]):
                    self.index = defaultindex
                else:
                    raise TypeError("Default index must be string or list of strings")

            self.search = [ ls.search for ls in logsource if ls.search ]

            self.conditions = [ ls.conditions for ls in logsource if ls.conditions ]        # build list of list of (field, value) tuples as base for merged query condition.
        elif type(logsource) == dict:       # create logsource configuration from parsed yaml
            self.merged = False
            if 'category' in logsource and type(logsource['category']) != str \
                    or 'product' in logsource and type(logsource['product']) != str \
                    or 'service' in logsource and type(logsource['service']) != str:
                raise SigmaConfigParseError("Logsource category, product or service must be a string")
            try:
                self.category = logsource['category']
            except KeyError:
                self.category = None
            try:
                self.product = logsource['product']
            except KeyError:
                self.product = None
            try:
                self.service = logsource['service']
            except KeyError:
                self.service = None
            if self.category == None and self.product == None and self.service == None:
                raise SigmaConfigParseError("Log source definition will not match")

            try:
                if type(logsource['rewrite']) is not dict:
                    raise SigmaConfigParseError("Rewrite rule must be a map")
                rewrite = logsource['rewrite']
                if not { 'category', 'product', 'service' }.issuperset(rewrite.keys()):
                    raise SigmaConfigParseError("Rewrite rule in log source configuration may only contain the keys 'category', 'product' and 'service'")
                if { str } != { type(value) for value in rewrite.values() }:
                    raise SigmaConfigParseError("Rewrite rule values may only contain strings")
                self.rewrite = tuple((rewrite.get(key) for key in ( 'category', 'product', 'service' )))    # build a (category, product, service) tuple from dict
            except KeyError:
                self.rewrite = None

            if 'index' in logsource:
                index = logsource['index']
                if type(index) not in (str, list):
                    raise SigmaConfigParseError("Logsource index must be string or list of strings")
                if type(index) == list and not all([type(index) == str for index in logsource['index']]):
                    raise SigmaConfigParseError("Logsource index patterns must be strings")
                if type(index) == list:
                    self.index = index
                else:
                    self.index = [ index ]
            else:
                # no default index handling here - this branch is executed if log source definitions are parsed from
                # config and these must not necessarily contain an index definition. A valid index may later be result
                # from a merge, where default index handling applies.
                self.index = []

            # free-form text search as-is. Instead of key=value, the value of 'search' is directly the value itself. Appended to conditions.
            search = logsource.get('search', None)
            if search not in [None, '']:
                if type(search) != str:
                    raise SigmaConfigParseError("search field must be a string: " + repr(search))
                self.search.append(logsource.get('search', ''))

            try:
                if type(logsource['conditions']) != dict:
                    raise SigmaConfigParseError("Logsource conditions must be a map")
                self.conditions = [ (field, value) for field, value in logsource['conditions'].items() ]    # build list of (field, value) tuples as base for query condition
            except KeyError:
                self.conditions = list()
        else:
            raise SigmaConfigParseError("Logsource definitions must be maps")

    def matches(self, category, product, service):
        """Match log source definition against given criteria, None = ignore"""
        searched = 0
        for searchval, selfval in zip((category, product, service), (self.category, self.product, self.service)):
            if searchval == None and selfval != None:
                return False
            if selfval != None:
                searched += 1
                if searchval != selfval:
                    return False
        if searched:
            return True

    def __str__(self):  # pragma: no cover
        return "[ LogSourceConfiguration: %s %s %s indices: %s ]" % (self.category, self.product, self.service, str(self.index))



# Output backends for sigmac
# Copyright 2016-2018 Thomas Patzke, Florian Roth (Nextron Systems), Roey

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import re

from sigma.parser.modifiers.type import SigmaRegularExpressionModifier

from sigma.parser.condition import SigmaAggregationParser
from .base import SingleTextQueryBackend
from .mixins import MultiRuleOutputMixin

class HumioBackend(SingleTextQueryBackend):
    """Converts Sigma rule into Humio query. Contributed by SOC Prime. https://socprime.com"""
    identifier = "humio"
    active = True

    reEscape = re.compile('("|(?<!\\\\)\\\\(?![*?\\\\]))')
    reClear = None
    andToken = " "
    orToken = " or "
    notToken = "!"
    subExpression = "%s"
    listExpression = "%s"
    listSeparator = " "
    valueExpression = "\"%s\""
    nullExpression = "NOT %s=\"*\""
    notNullExpression = "%s=\"*\""
    mapExpression = "%s=%s"
    regexExpression = "regex(\"%s=(\\\"%s\\\")\")"
    mapListsSpecialHandling = True
    mapListValueExpression = "%s IN %s"
    typedValueExpression = {
        SigmaRegularExpressionModifier: "/%s/"
    }

    def generateMapItemNode(self, node):
        key, value = node
        if isinstance(value, SigmaRegularExpressionModifier):# or isinstance(value, str) and "*" in value :
            return self.regexExpression % (key, self.cleanValue(value))
        else:
            return super().generateMapItemNode(node)

    def generateNOTNode(self, node):
        generated = self.generateNode(node.item)
        if generated is not None:
            return "%s(%s)" % (self.notToken, generated)
        else:
            return None

    def generateANDNode(self, node):
        generated = [self.generateNode(val) for val in node]
        filtered = [g for g in generated if g is not None]
        if filtered:
            if self.sort_condition_lists:
                filtered = sorted(filtered)
            if any([item for item in filtered if "regex" in item]):
                res = ""
                for item in filtered:
                    if item.startswith("regex"):
                        if res.endswith(" | "):
                            res = res.rstrip(" | ")
                        res += " | %s | " % item.strip(" | ")
                    else:
                        res += item
                return res.strip(" | ")
            return self.andToken.join(filtered)
        else:
            return None

    def generateORNode(self, node):
        generated = [self.generateNode(val) for val in node]
        filtered = [g.strip(" | ") for g in generated if g is not None]
        if filtered:
            if self.sort_condition_lists:
                filtered = sorted(filtered)
            if any([item for item in filtered if "regex" in item]):
                res = ""
                for item in filtered:
                    if item.startswith("regex"):
                        if res.endswith(" | "):
                            res = res.rstrip(" | ")
                        res += " | %s | " % item.strip(" | ")
                    else:
                        res += item
                return res.strip(" | ")
            return self.orToken.join(filtered)
        else:
            return None

    def cleanValue(self, val):
        if isinstance(val, SigmaRegularExpressionModifier):
            val = val.value
            if "\\" in val:
                val = re.sub(r"\\", r"\\\\\\", val)
        # if (val.startswith("*") or val.endswith("*")) and "\\" in val:
        #     val = re.sub(r"\\", r"\\\\\\", val)
        return super().cleanValue(val)

    def generateMapItemListNode(self, key, value):
        if isinstance(value, SigmaRegularExpressionModifier):
            key_mapped = self.fieldNameMapping(key, value)
            return {'regexp': {key_mapped: str(value)}}
        # if any([item for item in value if "*" in item]):
        #     return (" | " + " | ".join([self.regexExpression % (key, self.cleanValue(item)) for item in value]) + " | ")
        if not set([type(val) for val in value]).issubset({str, int}):
            raise TypeError("List values must be strings or numbers")
        return " or ".join(['%s=%s' % (key, self.generateValueNode(item)) for item in value])

    def generateAggregation(self, agg):
        if agg is None:
            return ""
        if agg.aggfunc == SigmaAggregationParser.AGGFUNC_NEAR:
            raise NotImplementedError("The 'near' aggregation operator is not yet implemented for this backend")
        if agg.groupfield is None:
            if agg.aggfunc_notrans == 'count':
                if agg.aggfield is None :
                    return " | val := count() | val %s %s" % (agg.cond_op, agg.condition)
                else:
                    agg.aggfunc_notrans = 'dc'
            return " | count(field=%s, distinct=true, as=val) | val %s %s" % (agg.aggfield or "", agg.cond_op, agg.condition)
        else:
            if agg.aggfunc_notrans == 'count':
                if agg.aggfield is None :
                    return " | val := count(field=%s) | val %s %s" % (agg.groupfield or "", agg.cond_op, agg.condition)
                else:
                    agg.aggfunc_notrans = 'dc'
            return " | groupby(field=%s, function=count(field=%s, distinct=true, as=val)) | val %s %s" % (agg.groupfield or "", agg.aggfield or "", agg.cond_op, agg.condition)

        
    def generate(self, sigmaparser):
        """Method is called for each sigma rule and receives the parsed rule (SigmaParser)"""

        for parsed in sigmaparser.condparsed:
            query = self.generateQuery(parsed)
            #before = self.generateBefore(parsed)
            #after = self.generateAfter(parsed)

            result = ""
            # if before is not None:
            #     result = before
            if query is not None:
                result += query
            # if after is not None:
            #     result += after
            if result.endswith(" | "):
                result = result.strip(" | ")
            return result



# Sigma configuration discovery
# Copyright 2016-2018 Thomas Patzke, Florian Roth

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from collections.abc import Iterable
from pathlib import Path
import sys
import re
from sigma.configuration import SigmaConfiguration
from sigma.config.exceptions import SigmaConfigParseError

class SigmaConfigurationManager(object):
    """
    Locate Sigma configuration files in a directory and provide them as well as information
    about them.
    """
    re_identifier = re.compile("^[\\w-]+$")
    def __init__(self, paths=None):
        """
        Initialize configuration collection. If paths is not given, some default locations are used:

        * Directory config/ in current script run path
        * 'data_files' sigma config location
        * ~/.config/sigma
        * /etc/sigma

        Parameters:
        * paths: list of strings with paths
        """
        if paths is None:
            self.paths = [
                    path for path in (
                            Path(sys.path[0]) / "config",           # Script launch directory + config/
                            Path(sys.prefix) / "local/etc/sigma",   # System location of installed pip package
                            Path(sys.prefix) / "etc/sigma",        # Virtualenv location of installed pip package
                            Path.home() / ".config/sigma",          # $HOME + .config/sigma
                            Path("/etc/sigma"),
                        )
                        if path.exists()
                    ]
        elif isinstance(paths, Iterable) and all([type(path) is str for path in paths]):
            self.paths = [Path(path) for path in paths]
        else:
            raise TypeError("None or iterable of strings expected as paths")

        self.configs = dict()
        self.errors = list()
        self.update()

    def update(self):
        """Update configurations"""
        self.configs.clear()
        self.errors.clear()
        for path in reversed(self.paths):       # Configs from first paths override latter ones
            for conf_path in path.glob("**/*.yml"):
                try:
                    f = conf_path.open()
                    self.configs[conf_path.stem] = SigmaConfiguration(f)
                    f.close()
                except (SigmaConfigParseError, OSError) as e:
                    self.errors.append((conf_path, e))

    def list(self):
        """Returns a list of (identifier, title) tuples of found configurations."""
        return [ (conf_id, config.config.setdefault("title", ""), config.config.setdefault("backends", list())) for conf_id, config in self.configs.items() ]

    def get(self, name):
        """
        Return a config by identifier or file path. First, it tries to resolve identifier from
        discovered configurations (file name stem). If this fails, the parameter value is treated
        as file name.
        """
        try:                # Lookup in discovered configurations
            return self.configs[name]
        except KeyError:    # identifier not found, try with filename
            f = open(name)
            return SigmaConfiguration(f)
